########################################################################
""" channel.py  -- Network channel w/ basic integrity checks

Copyright (c) 1999-2012, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel (sdeibel@archaeopteryx.com) and
John Ehresman (jpe@archaeopteryx.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Notes:

- User should be sure to destruct the CNetworkChannel object in
  all cases.  If this isn't done, the port may be left in use
  even after the application quits (depending on your OS)

"""
#########################################################################

import sys
import os
import errno
import socket

if sys.platform != 'win32':
  import fcntl
  try:
    if sys.hexversion >= 0x02020000:
      from fcntl import F_SETFD
    else:
      from FCNTL import F_SETFD
  except ImportError:
    # F_SETFD can't be imported on at least some python 2.0 builds
    F_SETFD = None

try:
  import ctypes
  if sys.platform == 'win32':
    import ctypes.wintypes  # Needs to be explicitly imported
except ImportError:
  ctypes = None

if sys.hexversion < 0x02020000:
  import whrandom
  randint = whrandom.randint
else:
  import random
  randint = random.randint

import pyutils

import struct
if sys.hexversion >= 0x03000000:
  import pickle as cPickle
  from io import BytesIO
else:
  import cPickle
import select
import time

ENABLE_LOGGING = 0

# This gets reset as soon as the dbgtracermodule is loaded to use
# its implementation (so we can get pid even in execution jails)
getpid = os.getpid

def doprint(*args):
  if not ENABLE_LOGGING:
    return
  txt = ' '.join([str(a) for a in args])
  dtime = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(time.time()))
  sys.stderr.write("wingdbg - %s - %s - CHANNEL: %s\n" % (dtime, getpid(), txt))
  
PROTOCOL = 1

#########################################################################
# Constants
#########################################################################

_kPacketIntSize = struct.calcsize('!I')

#-----------------------------------------------------------------------
def DisableSocketInheritance(sock):
  """ Disable socket inheritance into subprocesses.  Needed on all
  platforms -- on win32 sockets are set to inherit by default even
  though most other handle types are not; see
  http://support.microsoft.com/kb/150523 for confirmation.  This
  is used in debug server so wingutils.win32 is not used.  ctypes
  is used directly instead; function fails if ctypes is not available
  and returns 0, it returns 1 if it succeeds.
  """

  # Protect against failures, including old / broken ctypes installs
  try:
    if sys.platform != 'win32':
      if F_SETFD is None:
        return 0
      fcntl.fcntl(sock.fileno(), F_SETFD, 1)
      return 1
    
    # Rest is for win32; prep ctypes function here as this is not 
    # performance critical
    if ctypes is None:
      return 0
  
    kernel32 = ctypes.WinDLL('kernel32')
    SetHandleInformation = kernel32.SetHandleInformation
    SetHandleInformation.restype = ctypes.wintypes.BOOL
    SetHandleInformation.argtypes = [
      ctypes.wintypes.HANDLE,  #__in  HANDLE hObject,
      ctypes.wintypes.DWORD,  #__in  DWORD dwMask,
      ctypes.wintypes.DWORD,  #__in  DWORD dwFlags
    ]
  
    HANDLE_FLAG_INHERIT = 0x0001
    r = SetHandleInformation(sock.fileno(), HANDLE_FLAG_INHERIT, 0)
    if r == 0:
      raise ctypes.WinError()
    
    return 1
  except:
    # XXX Maybe should log to somewhere
    return 0

#-----------------------------------------------------------------------
# Utility to safely handle retrying select when EINTR signal is received
# (this happens whenever a signal of any kind comes in during select())
# Any exception other than EINTR is passed out by this call.
def safe_select(rsockets, wsockets, esockets, secs = None):
  
  EINTR = 4

  # Determine if there is an end_time
  if secs is None or float(secs) == 0.0:
    end_time = None
    secs = 0.0
  else:
    end_time = time.time() + secs
    
  #doprint('in select; secs =', secs)
  #import traceback
  #traceback.print_stack()
  
  # We loop until we stop getting EINTR, but not longer than the allocated timeout
  while end_time is None or time.time() < end_time:
    
    # Python 3.3 reworks IO errors so select.error is a deprecated alias for
    # OSError and apparently InterruptedError is the new EINTR
    if sys.hexversion >= 0x03030000:
      try:
        return select.select(rsockets, wsockets, esockets, float(secs))
      except InterruptedError:
        pass
      
    # Python 3.2 and earlier use select.error for EINTR
    else:
      try:
        return select.select(rsockets, wsockets, esockets, float(secs))
      except select.error:
        etype, val, tb = sys.exc_info()
        # Getting the errno from socket.error instance by python version:
        #   < 2.6 -- val[0]
        #   >= 2.6 and < 3.0 -- val[0] or val.errno
        #   >= 3.0 -- val.errno
        if sys.version_info < (3, 0):
          val_errno = val[0]
        else:
          val_errno = val.errno
        doprint("EXCEPTION IN SAFE SELECT: errno =", val_errno)
        if val_errno != EINTR:
          raise

    if end_time is not None:
      secs = max(0.0, end_time - time.time())
    else:
      break

  # Ran out of time:  So nothing is available
  return [], [], []

def IsWouldBlockExc(exc):
  """ Returns whether the exception indicates that the operation would
  block. """
  
  # The winsock api raises an exception w/ errno 10035 when send is called
  # on a socket and nothing can be written 
  if sys.platform == 'win32':
    WSAEWOULDBLOCK = 10035
    if exc.errno == WSAEWOULDBLOCK:
      return 1

  return exc.errno in (errno.EWOULDBLOCK, errno.EAGAIN)

def BlockingSend(sock, data):
  """ Send the data to the socket, not returning until everything is sent
  even if the socket is in nonblocking mode """
  
  total_count = 0
  remaining = data
  sent = 0
  
  while total_count < len(data):
    if sent != 0:
      remaining = remaining[sent:]
      
    try:
      sent = sock.send(remaining)
    except socket.error:
      #doprint('Exception when sending')
      etype, exc, tb = sys.exc_info()
      if not IsWouldBlockExc(exc):
        raise
      sent = 0
      
    if sent == 0:
      time.sleep(0.0001)
    else:
      total_count = total_count + sent

  return total_count


#########################################################################
# Safe unpickling of values w/o instantiating instances
#########################################################################

if sys.hexversion >= 0x03000000:
  class CSafeUnpickler(cPickle.Unpickler):
    def find_class(self, module, name):
      raise cPickle.UnpicklingError("global '%s.%s' is forbidden" %
                                     (module, name))
  def GetSafeUnpickler(packetdata):
    f = BytesIO(packetdata)
    return CSafeUnpickler(f)
  
else:
  def GetSafeUnpickler(packetdata):
    f = pyutils.CStringIO(packetdata)
    p = cPickle.Unpickler(f)
    p.find_global = None
    return p
  
class CDefaultSerializer:
  """ Serialization using pickle, no class instances allowed. """
  
  def Deserialize(self, bytes_data):
    
    # When possible (Python 2.2+), prevent loading of any instances
    if sys.hexversion < 0x02020000:
      obj = cPickle.loads(bytes_data)
    else:
      p = GetSafeUnpickler(bytes_data)
      obj = p.load()
      
    return obj
  
  def Serialize(self, obj):
    
    return cPickle.dumps(obj, PROTOCOL)
    
    
#########################################################################
# Classes
#########################################################################

class CNetworkChannel:
  """ Simple class for managing a network connection with two-way Python data
  transmission. This code requires availability of previously entered matching
  "secret" keys at each end of the channel. The channel is established by a
  client that starts by listening on a socket and requesting (externally) that
  a server connect back to it. The server then generates a packet sequence
  seed number. The client responds with its own packet sequence number for
  packets sent from server back to client. Subsequently, packets must be
  received according to previously determined ever-increasing sequence
  numbers. Any packets out of order are rejected.

  Two special initiation packets are sent to open the channel. These are
  packaged by the cPickle module. The server in connecting back to the client
  first sends the following:

       int     -- Code version number for the server
       int     -- Sequence number for packets that get sent from client
                  to server

  The client then responds with a packet of the same structure, with its
  own code version number and random packet seed.

  Subsequent packets sent in either direction are sequence numbered. On the
  wire, these packets are sent each with prepended length information as
  follows:

      int      -- Length of rest of packet
      string   -- Packet data

  The packet length value is built using the Python 'struct' module, so
  that it is always a Python int and can be read from the socket as a
  known finite length value before the rest of the variable length packet
  is read.

  The data portion of that contains a tuple with the following, packed
  up with cPickle:

      int      -- Packet sequence number
      string   -- User's data

  cPickle takes care of byte order issues, etc, so communications between 
  mixed architectures should work.

  """

  # Current version number of this code
  kCodeVersion = 2

  # Exception classes
  class CBadParam(Exception):
    def __init__(self, description):
      Exception.__init__(self, description)
      self.descr = description
    def __repr__(self):
      return repr(self.descr)
    
  class CBadPacket(Exception):
    def __init__(self, description):
      Exception.__init__(self, description)
      self.descr = description
    def __repr__(self):
      return repr(self.descr)
    
  class CSocketError(Exception):
    def __init__(self, id, description):
      Exception.__init__(self, description)
      self.id = id
      self.descr = description
    def __repr__(self):
      return repr((self.id, self.descr))

  class CNetworkTimeout(Exception):
    def __init__(self, description):
      Exception.__init__(self, description)
      self.description = description
    def __repr__(self):
      return self.description

  # A default serializer that's used if no other serializer is specified
  _fSerializer = CDefaultSerializer()

  #-----------------------------------------------------------------------
  def __init__(self, timeout, serializer=None):
    """ Constructor: User must subsequently call either Connect() or Listen()
    to start up the channel. Timeout is network timeout in seconds, to
    terminate suspiciously silent connections (this is used only within the
    packet body; a blocking read from packet never times out).

    """
    
    # Clear/init
    self.__fHost = None
    self.__fPort = None
    self.__fNetworkTimeout = timeout
    self.__fLocalSequence = randint(0, 0x0fffffff)
    self.__fRemoteSequence = 0
    self.__fLocalVersion = self.kCodeVersion
    self.__fRemoteVersion = 0
    self._fSocket = None
    self._fPeerIP = None
    self.__fCloseCallback = None
    self.__fForkCallback = None
    self.fListening = 0
    
    if serializer:
      self._fSerializer = serializer
    
  #-----------------------------------------------------------------------
  def __del__(self):
    """ Destructor: Make sure we close the network connection """
 
    self.Close()

  #-----------------------------------------------------------------------
  def Connect(self, host, port):
    """ Connect to the other end of the channel. This will also send the
    initial packet that establishes the channel. """

    def is_ip_addr(name):
      t = name.split('.')
      if len(t) != 4:
        return 0
      for i in t:
        try:
          ii = int(i)
        except ValueError:
          return 0
        if ii < 0 or ii > 255:
          return 0
      return 1
        
    # Save values
    self.__fHost = host
    self.__fPort = port

    # Set up network connection back to remote client
    self._fSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    DisableSocketInheritance(self._fSocket)
    if sys.hexversion >= 0x02000000: # Works in >= 2.0.0 only    
      self._fSocket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    self._fSocket.connect((host, int(port)))
    # Avoid calling DNS unless needed, to avoid problems on hosts
    # with misconfigured DNS
    # Note: Changing this to getpeername() on socket may break on some systems
    if is_ip_addr(host):
      self._fPeerIP = host
    else:
      self._fPeerIP = socket.gethostbyname(host)

    # Initial handshake is started by connector, and finished by client
    self._WriteInitiationPacket()
    self._ReadInitiationPacket()

  #-----------------------------------------------------------------------
  def Listen(self, valid_ip_seq, host, port, blocking=1, keep_listening=False,
             listen_queue_size=1):
    """ Listen for connection back from the given server. When blocking == 1,
    this will block until the connection occurs so you'll need to run it in a
    thread where that is acceptible. Otherwise, it checks for a connection
    request and exits if there isn't one. The call returns the CNetworkChannel
    for the accepted connection (this is self if keep_listening is False, or
    otherwise a new instance) or None if no connection was accepted. This call
    may raise an exception in some cases, such as when the connecting machine
    doesn't match the given server host. Note that host may be None or '' to
    specify listening on all network interfaces. If the ip # of the incoming
    connection is not in the valid_ip_seq sequence, an CBadPacket exception is
    thrown. Elements of valid_ip_seq must be strings in dotted-quad notation.
    """

    # Create listening socket if not already there
    if self._fSocket is None:
      self._fSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      
      DisableSocketInheritance(self._fSocket)
        
      # Set SO_REUSEADDR so socket can be reused w/o waiting for X (30+?)
      # seconds on posix; not needed on win32
      if sys.platform != 'win32':
        self._fSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

      if sys.hexversion >= 0x02000000: # Works in >= 2.0.0 only
        self._fSocket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
      if host is None:
        host = ""
      self._fSocket.bind((host, port))
      self._fSocket.listen(listen_queue_size)
      self.fListening = 1
   
    # If not blocking then check first for availability of a connection
    # request and exit if there isn't one
    if not blocking:
      iready, oready, eready = safe_select([self._fSocket,], [], [], 0.0)
      if len(iready) == 0:
        return None

    # Listen for the initial connection from given server
    conn, addr = self._fSocket.accept()
    
    # Check if connection is from a valid peer
    if addr[0] not in valid_ip_seq:

      # Do this only on failure of first test to avoid calling gethostbyname()
      # as often as possible to avoid problems on hosts with misconfigured DNS
      
      # Adjust valid ip sequence:  We add localhost (127.0.0.1) if the
      # machine's primary IP address is in the list, and we add the machine's
      # primary IP address if localhost (127.0.0.1) is in the list : These
      # are the same thing from a security standpoint and apparently
      # accept() will sometimes report one and sometimes the other for
      # the remote peer
      try:
        primary_ip = socket.gethostbyname(socket.gethostname())
      except socket.error:
        primary_ip = '127.0.0.1'
      if '127.0.0.1' in valid_ip_seq and not primary_ip in valid_ip_seq:
        valid_ip_seq.append(primary_ip)
      if primary_ip in valid_ip_seq and not '127.0.0.1' in valid_ip_seq:
        valid_ip_seq.append('127.0.0.1')
    
    # Try again to see if connection is from a valid peer
    if addr[0] not in valid_ip_seq:    
      self.Close()
      conn.close()
      conn = None
      doprint("INVALID HOST", addr[0], "NOT IN", valid_ip_seq)
      raise self.CBadPacket("Connection from invalid host!")
    
    # Valid connection: Either reuse this channel or create a new one, according
    if keep_listening:
      new_channel = CNetworkChannel(self.__fNetworkTimeout, self._fSerializer)
    else:
      self._fSocket.close()
      self.fListening = 0
      new_channel = self
    new_channel._fSocket = conn
    new_channel._fPeerIP = addr[0]
    DisableSocketInheritance(new_channel._fSocket)
    if sys.hexversion >= 0x02000000: # Works in >= 2.0.0 only
      new_channel._fSocket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    
    # Initiation handshake
    new_channel._ReadInitiationPacket()
    new_channel._WriteInitiationPacket()

    # Success
    return new_channel

  #----------------------------------------------------------------------
  def SendPacket(self, data):
    """ Send the given data on the channel."""

    packet = self.__MarshallPacket(data)
    self.__WriteRawPacket(packet)

  #----------------------------------------------------------------------
  def WaitForPacket(self, secs):
    """ Wait for incoming data for given number of seconds, expressed as
    a floating point number.  Returns 1 if any data is available. 
    Use this call to most efficiently control latency in your protocol:
    Call it first and then ReceivePacket(blocking=1), rather than 
    repeatedly polling ReceivePacket(blocking=0). """
    
    if self._CallForkCallback()[0]:
      return 0
    
    iready, oready, eready = safe_select([self._fSocket,], [], [], secs)
    if len(iready) > 0:
      return 1
    else:
      return 0

  #----------------------------------------------------------------------
  def ReceivePacket(self, blocking=1):
    """ Get next data off the channel.  This will block until
    data is available unless you send blocking=0.  In non-blocking mode,
    the read operation will return None if no data is ready.  However, it will
    read the entire packet if some data is ready, so this is not 
    strictly asynchronous for very large data transfers. """

    packetdata = self.__ReadRawPacket(blocking)
    data = self.__UnmarshallPacket(packetdata)
    return data

  #----------------------------------------------------------------------
  def SetCloseCallback(self, cb):
    """ Set callback to invoke when the socket is closed:  This is usually
    used to cleanly deregister the socket from some async polling framework """
    
    self.__fCloseCallback = cb
    
  #----------------------------------------------------------------------
  def SetForkCallback(self, cb):
    """ Set callback to invoke when a child process is forked"""
    
    self.__fForkCallback = cb
    
  #----------------------------------------------------------------------
  def Close(self):
    """ Close connection to remote peer if open """
    if self._fSocket is not None:
      if self.__fCloseCallback is not None:
        self.__fCloseCallback(self._fSocket)
        self.__fCloseCallback = None
      if self._fSocket is not None:
        self._fSocket.close()
      self._fSocket = None
      self._fPeerIP = None

  #----------------------------------------------------------------------
  def IsClosed(self):
    return self._fSocket is None

  #----------------------------------------------------------------------
  def GetSocket(self):
    """ Get the socket being used for this channel.  Should only be
    used for external polling of data on socket, e.g., with select().
    Reading or writing on the socket will likely crash the channel. """
    return self._fSocket

  #----------------------------------------------------------------------
  def GetPeerIP(self):
    """ Get peer's IP address """
    return self._fPeerIP
  
  #----------------------------------------------------------------------
  def _UseSocket(self, sock):
    """ Set the socket object to use for the channel:  This is used *only*
    to wrap the socket returned from GetSocket() with another object
    that may handle some aspects of async socket management. """
    
    self._fSocket = sock
    
  #----------------------------------------------------------------------
  def _CallForkCallback(self):
    """Calls the fork detection callback and return (forked, exited)
    to indicate whether a fork occurred and whether debugging exited."""

    if self.__fForkCallback:
      return self.__fForkCallback()
    
    return (0, 0)
    
  #----------------------------------------------------------------------
  def __WriteRawPacket(self, packetdata):
    """ Write a raw packet to the wire.  Adds only a leading packet
    length so reader knows how much to read.  The packet data is
    assumed to be already packed in network byte order and ready 
    for network transfer. """

    # Error check
    if self._fSocket is None:
      raise self.CBadParam("No connection")

    # Check if we're a child process although this check is a too late if
    # debugging child processes and we try to cover most cases in netserver.py
    # so that the channel can be replaced before trying to write packets
    if self._CallForkCallback()[0]:
      return
    
    # Write data
    doprint("Sending packet length", len(packetdata))
    packet = struct.pack('!I', len(packetdata))
    try:
      BlockingSend(self._fSocket, packet)
      BlockingSend(self._fSocket, packetdata)
    except socket.error:
      etype, exc, tb = sys.exc_info()
      try:
        eid, descr = exc
      except:
        descr = exc
        eid = -1
      raise self.CSocketError(eid, descr)
    except:
#      doprint("EXCEPTION in __WriteRawPacket()")
#      doprint(sys.exc_info())
      raise

  #----------------------------------------------------------------------
  def __ReadRawPacket(self, blocking=1):
    """ Read a raw packet from the wire.  Will block until there is data
    available unless you send blocking=0.  In non-blocking mode, the read
    operation will return None if no data is ready.  However, it will read
    the entire packet if some data is ready, so this is not strictly
    asynchronous for very large data transfers. """

    # Error check
    if self._fSocket is None:
      raise self.CBadParam("No connection")

    # Check if we're a child process although this check is a too late if
    # debugging child processes and we try to cover most cases in netserver.py
    # so that the channel can be replaced before trying to read packets
    if self._CallForkCallback()[0]:
      return None
    
    # Read in non-blocking mode: trap exception indicating not-ready
    if not blocking:

      try:
#        peer = self.__fSocket.getpeername()
#        doprint("Socket peer in __ReadRawPacket", peer)
        iready, oready, eready = safe_select([self._fSocket,], [], [], 0.0)
#        doprint(iready, oready, eready)
        if len(iready) == 0:
          packetdata = None
        else:
          packetdata = self.__InternalReadRawPacket()
      except:
#        doprint("EXCEPTION in __ReadRawPacket()")
#        doprint(sys.exc_info())
        nt, val, tb = sys.exc_info()
        if nt == socket.error:
          try:
            id, descr = val
          except:
            descr = val
          raise self.CSocketError(id, descr)
        else:
          raise  # Reraises current exception

    # Read in blocking mode: any exception propagates outward
    else:

      # Wait forever for data on socket
      iready, oready, eready = safe_select([self._fSocket,], [], [self._fSocket,], 0.5)
      while len(iready) == 0 and len(eready) == 0 and not self.IsClosed():
        iready, oready, eready = safe_select([self._fSocket,], [], [self._fSocket,], 0.5)

      # Read it when it arrives
      if len(iready) > 0:
        packetdata = self.__InternalReadRawPacket()

      # Socket error or closed
      else:
        packetdata = None

    # Return the data
    return packetdata

  #----------------------------------------------------------------------
  def __CanRead(self, abs_timeout = 0):
    """ Returns whether there is something to read in the socket.  abs_timeout is
    the time index at which to stop selecting,  If it's 0 or < time.time(), select
    does not wait. """

    if abs_timeout <= 0:
      timeout = 0
    else:
      timeout = max(abs_timeout - time.time(), 0)
    iready, oready, eready = [self._fSocket,], [], [self._fSocket,]
    iready, oready, eready = safe_select(iready, oready, eready, timeout)
    if len(eready) != 0:
      raise self.CSocketError(32, "Tcp/ip exception detected")
    
    return (len(iready) != 0)
  
  #----------------------------------------------------------------------
  def __ReadBytes(self, n, abs_timeout):
    """ Reads n bytes from the socket.  Throws timeout exception if time.time()
    exceeds abs_timeout. """

    parts = []
    total_len = 0
    while total_len < n:
      ready = self.__CanRead(abs_timeout)
      if not ready:
        raise self.CNetworkTimeout("Network timeout while reading packet")
      
      try:
        # Never try to read more than 5K at one time (purposely small to keep
        # the GUI responsive)
        max_read = min(n - total_len, 5 * 1024)
        data = self._fSocket.recv(max_read)
        #doprint('read', repr(data))
        if len(data) == 0:
          self.Close()
          raise self.CSocketError(32, "Broken Pipe")
        parts.append(data)
        total_len = total_len + len(data)
        
      except socket.error:
        t, v, tb = sys.exc_info()
        if sys.platform == 'win32':
          if v[0] == 10035:         # Resource temporarily unavailable (so we try again)
            pass
          else:
            raise
        else:
          if v[0] == errno.EAGAIN:  # Resource temporarily unavailable (so we try again)
            pass
          else:
            raise
          
      except:
        #doprint('unexpected exception')
        #import traceback
        #traceback.print_exc()
        raise
      
    return pyutils.datajoin(parts)
  
  #----------------------------------------------------------------------
  def __InternalReadRawPacket(self):
    """ Read a raw packet from the wire with blocking behaviour as already
    set in the socket.  This is a utility that should only be called from
    __ReadRawPacket(). """

    # Error check
    if self._fSocket is None:
      raise self.CBadParam("No connection")

    # Set up for timeout
    timeout = time.time() + self.__fNetworkTimeout;

    # First read packet length
    packetdata = self.__ReadBytes(_kPacketIntSize, timeout)
    packetlen, = struct.unpack('!I', packetdata)
    packetlen = int(packetlen)
    doprint("Reading packet length", packetlen)

    # Then read rest of packet
    packetdata = self.__ReadBytes(packetlen, timeout)

    # Return the data
    return packetdata

  #-----------------------------------------------------------------------
  def _WriteInitiationPacket(self):
    """ Write handshake initiation packet containing code version
    information and random packet sequence number seed."""

    # Pack up and send rest of packet
    initiation = (self.__fLocalVersion, self.__fLocalSequence)
    doprint("Initiation packet contents sent:", initiation)
    packet = self._fSerializer.Serialize(initiation)
    self.__WriteRawPacket(packet)

  #-----------------------------------------------------------------------
  def _ReadInitiationPacket(self, blocking=1):
    """ Read handshake initiation packet containing code version
    information and random packet sequence number seed.  Returns whether
    packet was read."""

    # Get the packet and unpack common data
    data = self.__ReadRawPacket(blocking)
    if data is None:
      return 0
    
    initiation = self._fSerializer.Deserialize(data)
    doprint("Initiation packet contents read:", initiation)
    self.__fRemoteVersion, self.__fRemoteSequence = initiation
    return 1

  #----------------------------------------------------------------------
  def __MarshallPacket(self, data):
    """ Marshall a session packet. """  

    # Pack up data and sequence number
    #doprint("marshalling data", repr(data))
    packetdata = self._fSerializer.Serialize((self.__fRemoteSequence, data))
    self.__NextRemoteSequenceNumber()
    #doprint("Marshalled byte count:", len(packetdata))
    return packetdata
 
  #----------------------------------------------------------------------
  def __UnmarshallPacket(self, packetdata):
    """ Unmarshall the given packet.  Returns data or raises an exception
    if the packet was bad. """

    if packetdata is None:
      return None

    # Extract sequence number and user data.
    #doprint("Unmarshall byte count:", len(packetdata))
    sequence, userdata = self._fSerializer.Deserialize(packetdata)
      
    # Check validity
    if sequence != self.__fLocalSequence:
      self.Close()
      raise self.CBadPacket('ALERT: Bad packet sequence number')
    else:
      self.__NextLocalSequenceNumber()

    # Return value
    #doprint("unmarshalled data", repr(userdata))
    return userdata

  #----------------------------------------------------------------------
  def __NextLocalSequenceNumber(self):
    """ Bump local packet sequence number to next available value """
    if self.__fLocalSequence == 0x0fffffff:
      self.__fLocalSequence = 0
    else:
      self.__fLocalSequence = self.__fLocalSequence + 1

  #----------------------------------------------------------------------
  def __NextRemoteSequenceNumber(self):
    """ Bump remote packet sequence number to next available value """
    if self.__fRemoteSequence == 0x0fffffff:
      self.__fRemoteSequence = 0
    else:
      self.__fRemoteSequence = self.__fRemoteSequence + 1


#########################################################################
# Test code
#########################################################################

# To test, start a client first and then the server
if __name__ == '__main__':

  HOST = '127.0.0.1'        # Symbolic name meaning the local host
  PORT = 50006              # Arbitrary non-privileged server
  BLOCKING = 0              # Test async connection

  # Run as server
  if len(sys.argv) > 1 and sys.argv[1] == 'server':
    doprint("Server connecting back to client...")
    channel = CNetworkChannel(30)
    channel.Connect(HOST, PORT)
    doprint("Server connection established")
    doprint("Server sending test values")
    testpacket = ( 145, 38192, 'This is a test', {} )
    doprint(testpacket)
    channel.SendPacket(testpacket)
    doprint("Server waiting for reply")
    testpacket2 = channel.ReceivePacket()
    doprint("Server received", testpacket2)
    if testpacket != testpacket2:
      doprint("MISMATCHED VALUES RETURNED!")
    doprint("Server done")
    channel = None

  # Run as client
  else:
    channel = CNetworkChannel(30)
    doprint("Client listening..."    )
    while channel.Listen([HOST,], HOST, PORT, blocking = BLOCKING) == 0:
      time.sleep(0.5)
    doprint("Client connection established")
    testpacket = channel.ReceivePacket()
    doprint("Client received", testpacket)
    doprint("Client sending values back")
    channel.SendPacket(testpacket)
    doprint("Client done")
    channel = None
