#########################################################################
""" callutils.py -- Utilities for dealing with Python callable objects

Copyright (c) 1999-2012, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel and John P. Ehresman

"""
#########################################################################

import types
import functools
import copy

import reflect

#-----------------------------------------------------------------------
def countargs(clb):
  """ Utility to count number of args for given callable object """

  assert callable(clb), "Not a callable object"

  if type(clb) is types.FunctionType: # aka LambdaType
    return clb.func_code.co_argcount
  elif type(clb) is types.MethodType:
    return clb.im_func.func_code.co_argcount
  elif type(clb) is types.ClassType:
    return clb.__init__.im_func.func_code.co_argcount
  elif type(clb) is types.InstanceType and hasattr(clb, '__call__'):
    return clb.__call__.im_func.func_code.co_argcount
  else:
    assert 0, "Unhandled callable type"

#-----------------------------------------------------------------------
def flexapply(clb, *args):
  """ Apply a callable to a list of args.  This is like apply() except
  that the given args are truncated if too long for the function and
  extended with None entries if too short """

  # Build matching args list
  nargs = countargs(clb)
  pargs = []
  for i in range(0, nargs-1):
    try:
      pargs.append(args[i])
    except:
      pargs.append(None)

  # Make the call
  return apply(clb, pargs)


#######################################################################
class log_exceptions(object):
  """ Decorator class that will catch exceptions from the function it
  wraps and return the given value. """
  
  # Change the following to True to log even in debug mode
  log_in_debug_mode = False
  
  def __init__(self, exc_return_val, exc_call=None):
    """ Create the decorator.  exc_return_val is the value to return
    on exception and exc_call will be called with original arguments to
    wrapped function if not None. """
    
    self.exc_return_val = exc_return_val
    self.exc_call = exc_call    
    
  def __call__(self, func):
    """ Excecuted when decorator applied to wrapped function. """
    
    if __debug__:
      if not self.log_in_debug_mode:
        return func

    # Bind vals to locals so self can disappear
    exc_return_val = self.exc_return_val
    exc_call = self.exc_call
    
    def wrapper(*args, **kw):
      try:
        return func(*args, **kw)
      except Exception:
        reflect.ReportCurrentException()

        if exc_call is not None:
          try:
            exc_call(*args, **kw)
          except Exception:
            reflect.ReportCurrentException()
          
        return copy.deepcopy(exc_return_val)
      
    try:
      functools.update_wrapper(wrapper, func)
    except AttributeError:
      pass # Happens with methods on cython cdef class 
    
    return wrapper
        
        
def profile(callable):
    """ Decorator to profile a single callable and print results to stdout. """
    
    def cprofile_wrapper(*args, **kw):
        import cProfile, pstats
        
        prof = cProfile.Profile()
        try:
          return prof.runcall(callable, *args, **kw)
        finally:
          print 'Profile completed'
          
          stats = pstats.Stats(prof)
          stats.strip_dirs()
          stats.sort_stats('cumulative', 'calls')
          stats.print_stats(40)    
          stats.sort_stats('time', 'calls')
          stats.print_stats(40)

    wrapper = cprofile_wrapper
    
    import functools
    functools.update_wrapper(wrapper, callable)
    return wrapper
