########################################################################
""" wingutils/destroyable.py -- Support for explicit instance destruction

Copyright (c) 1999-2015, Archaeopteryx Software, Inc.  All rights reserved.

Written by John P. Ehresman and Stephan R.A. Deibel

Notes:


"""
#########################################################################

import os
import collections
import types
import new
import logging

# Allow import to continue if Qt is unavailable since this is used at build time
try:
  from guiutils.qt_base import QObject
except ImportError:
  QObject = None
  
# Note that the _ prefixed names are not accessible from outside the modules when
# using the cython compiled version
_gEmissionStack = []
_gClassSignalHandlers = {}
_gSignalsForCls = {}
_kEmptyFrozenset = frozenset()
_gLogAllEmissions = False

def GetLogAllEmissions():
  return _gLogAllEmissions

def SetLogAllEmissions(log_all):
  global _gLogAllEmissions
  _gLogAllEmissions = log_all

class _CHandlerRecord:
  
  def __init__(self, signal_name, proc, extra, cls):
    """ Information for one connected handler -- cls will be set to a 
    class object if it's a class wide handler or cls will be None if it's
    a handler for a specific instance """
    
    self.signal_name = signal_name
    self.proc = proc
    self.extra = extra
    self.cls = cls

class _CEmissionStackItem:
  """ The item in the _gEmissionStack for one signal emission. """
  
  def __init__(self, obj, signal_name, handlers):
    
    self.obj = obj
    self.signal_name = signal_name
    self.handlers = handlers
    self.stopped = False
  
def _Disconnect(inst_dict, cls_dict, handler):
  """ Remove handler from either the instance or class dict and invalidate
  handler.  Returns whether handler was in one of the dicts """
  
  found = False
  if inst_dict is not None:
    found = _RemoveFromDict(inst_dict, handler)
  if not found:
    found = _RemoveFromDict(cls_dict, handler)

  _ClearHandler(handler)
  
  return found
    
def _ClearHandler(handler):
  
  handler.proc = None
  handler.extra = ()
  handler.cls = None  
  
def _Emit(obj, inst_handler_dict, signal_name, args):
  """ Call all handlers connected to the given signal using given
  tuple of args.  Returns True if the signal emission was stopped
  by a handler, or otherwise False."""
  
  handlers = _GetHandlers(obj, inst_handler_dict, signal_name)
  if not handlers:
    return False
  
  emission_item = _CEmissionStackItem(obj, signal_name, handlers)
  _gEmissionStack.append(emission_item)
  try:
    for single_handler in handlers:
      if single_handler.proc is not None:
        _EmitToSingleHandler(single_handler, args)
        if emission_item.stopped:
          return True
      
    return False
  finally:
    del _gEmissionStack[-1]
    
def _EmitStopByName(obj, signal_name):

  for emission_item in reversed(_gEmissionStack):
    if emission_item.obj == obj and emission_item.signal_name == signal_name:
      emission_item.stopped = True
      break
      
def _GetHandlers(obj, inst_handler_dict, signal_name):
  """ Get list of handlers for given signal name.  A new list is always
  returned """
  
  inst_handlers = (inst_handler_dict.get(signal_name) if inst_handler_dict is not None
                   else None)
  cls_handlers = _gClassSignalHandlers.get(signal_name)
  handlers = []
  if inst_handlers is not None:
    handlers.extend(inst_handlers)
  if cls_handlers is not None:
    for single_handler in cls_handlers:
      if isinstance(obj, single_handler.cls):
        handlers.append(single_handler)
    
  return handlers
  
def _EmitToSingleHandler(handler, args):
  """ Emit to one handler.  Takes care of appending any extra args. """
  
  if len(handler.extra) != 0:
    args = args + handler.extra
    
  proc_name = None
  if _gLogAllEmissions:
    proc_name = repr(handler.proc)
    print 'Calling handler for signal {}: {}'.format(handler.signal_name,
                                                     proc_name)
    
  try:
    handler.proc(*args)
  except:
    logging.exception("Error during emit(): %r" % handler.proc)
    
  if _gLogAllEmissions:
    print 'Handler finished: {}'.format(proc_name)

def _RemoveFromDict(d, handler):
  """ Remove given handler from dictionary.  Returns if the handler was
  found and removed. """
  
  handler_list = d.get(handler.signal_name)
  if handler_list is None:
    return False
  
  for i, handler_item in enumerate(handler_list):
    if handler_item is handler:
      del handler_list[i]
      if len(handler_list) == 0:
        del d[handler.signal_name]
      return True
      
  return False


#########################################################################
# Destroyable object base class
#########################################################################
def _GetSignalsForCls():
  return _gSignalsForCls

class CDestroyable(object):
  """ A class that implements a destroy method that calls all _destroy_impl
  methods defined in any defining class of the instance and clears the
  instance's dictionary.  The _destroy_impl methods will be called in an
  order such that a method defined in a derived class will be called before
  methods defined in a base class.  It will also emit a 'destroy' signal before
  any _destroy_impl method is called. """

  #----------------------------------------------------------------------
  def __init__(self, signals = ()):
    """ Constructor.  Asserts that self.destroy refs the CDestroyable.destroy
    method. """

    # Make sure signals are defined for class
    existing = _gSignalsForCls.get(type(self))
    if existing is None:
      _gSignalsForCls[type(self)] = set(list(signals) + ['destroy'])
    else:
      assert all(name in existing for name in existing)
      assert 'destroy' in existing
    
    # A map of signal_name keys to values which are lists of_CHandlerRecord's
    self.__fSignalCatchers = {}

    self.__fDestroyed = 0

  #----------------------------------------------------------------------
  def __del__(self):
    """ Delete method so that destroy is invoked if object is GC'd. """

    #print self.__dict__
    try:
      self.destroy()
    except:
      if __debug__:
        import traceback
        traceback.print_exc()
      else:
        raise

  #----------------------------------------------------------------------
  def destroy(self, *args):
    """ Default destroy method.  Calls _destroy_impl in each class in the 
    object's class & base classes """
    
    if self.destroyed():
      return

    self.__fDestroyed = True

    self.emit('destroy', self)

    def get_impl_list(self):
      assert not isinstance(self, types.ClassType)
      impl_list = []
      mro_list = type.mro(self.__class__)
      for cls in mro_list:
        
        impl = cls.__dict__.get('_destroy_impl')
        if impl is not None:
          impl_list.append(impl)
      return impl_list
          
    impl_list = get_impl_list(self)
    for impl in impl_list:
      try:
        impl(self)
      except:
        logging.exception("Exception raised while calling: %r" % (impl,))

    # Clear all instance attributes and then reset __fDestroyed
    try:
      d = self.__dict__
    except AttributeError:
      pass
    else:
      d.clear()
      
    self.__fSignalCatchers = None
    self.__fDestroyed = True

  #----------------------------------------------------------------------
  def destroyed(self):
    """ Returns boolean indicating whether self has been destroyed. """

    try:
      return self.__fDestroyed
    except AttributeError:
      # This may be invoked from a __del__ method that is called when
      # self.__dict__.clear() is invoked.  To handle this case, return True
      # iff self.__dict__ is empty.  Always returning True on AttributeError
      # hides exceptions raised because __init__ is never called
      #
      # XXX Currently disabled because there is code that sets an attribute
      # on an object after it is destroyed
      #
      #if len(self.__dict__) != 0:
      #  raise
      return True
      
  #----------------------------------------------------------------------
  @classmethod
  def class_connect(cls, signal_name, proc, *extra):
    """Connect handler for signal."""

    record = _CHandlerRecord(signal_name, proc, extra, cls)
    handler_list = _gClassSignalHandlers.get(signal_name)
    if handler_list is None:
      handler_list = _gClassSignalHandlers[signal_name] = []
    handler_list.append(record)
    
    return record

  #----------------------------------------------------------------------
  @classmethod
  def class_connect_while_alive(cls, signal_name, proc, alive_obj, *extra):
    """Connect handler for signal while alive_obj is not destroyed."""

    handler_id = cls.class_connect(signal_name, proc, *extra)

    if QObject is not None and isinstance(alive_obj, QObject):
      def cb(*args):
        cls.__class_disconnect_on_destroy(alive_obj, handler_id)
      alive_obj.destroyed.connect(cb)
    else:
      alive_obj.connect('destroy', cls.__class_disconnect_on_destroy, handler_id)

    return handler_id
    
  #----------------------------------------------------------------------
  @classmethod
  def class_disconnect(cls, handler):
    """Connect handler for signal while alive_obj is not destroyed."""

    _Disconnect(None, _gClassSignalHandlers, handler)

  #----------------------------------------------------------------------
  @classmethod
  def __class_disconnect_on_destroy(cls, other, handler):
    """Connect handler for signal while alive_obj is not destroyed."""

    _Disconnect(None, _gClassSignalHandlers, handler)
    
  #----------------------------------------------------------------------
  def connect(self, signal_name, proc, *extra):
    """Connect handler for signal."""

    if signal_name not in self.signals():
      raise TypeError, 'Bad signal %s passed to connect' % signal_name

    record = _CHandlerRecord(signal_name, proc, extra, None)
    handler_list = self.__fSignalCatchers.get(signal_name)
    if handler_list is None:
      handler_list = self.__fSignalCatchers[signal_name] = []
    handler_list.append(record)
    
    return record
  
  #----------------------------------------------------------------------
  def __disconnect_on_destroy(self, other, handler):
    """ Disconnect one of our handler when the other object is destroyed. """

    # Uncomment to see disconnections
    #print '__disconnect_on_destroy', self
    self.disconnect(handler)
    
  #----------------------------------------------------------------------
  def connect_while_alive(self, signal_name, proc, alive_obj, *extra):
    """Connect handler for signal while alive_obj is not destroyed."""

    if len(extra) == 0:
      id = self.connect(signal_name, proc)
    else:
      connect_args = (signal_name, proc) + extra
      id = apply(self.connect, connect_args)

    if QObject is not None and isinstance(alive_obj, QObject):
      def cb(*args):
        self.__disconnect_on_destroy(alive_obj, id)
      alive_obj.destroyed.connect(cb)
    else:
      alive_obj.connect('destroy', self.__disconnect_on_destroy, id)

    return id
  
  #----------------------------------------------------------------------
  def disconnect(self, record):
    """ Disconnect previously connected handler. """
    
    
    # None shouldn't be passed but check for it anyway since it can cause
    # the cython compiled version to segfault 
    assert record is not None
    if record is None:
      return

    if self.destroyed():
      _ClearHandler(record)
      return
    
    _Disconnect(self.__fSignalCatchers, _gClassSignalHandlers, record)

  #----------------------------------------------------------------------
  def disconnect_by_func(self, func):
    """ Disconnect previously connected handler. """

    if self.destroyed():
      return

    for d in [self.__fSignalCatchers, _gClassSignalHandlers]:
      for signal_name, catcher_list in d.items():
        for i, catcher in enumerate(catcher_list):
          if catcher.proc == func:
            del catcher_list[i]
            if len(catcher_list) == 0:
              del d[signal_name]

            catcher.proc = None
            catcher.extra = ()
            catcher.cls = None
            return
          
  #----------------------------------------------------------------------
  def emit(self, signal_name, *args):
    """ Emit the given signal with args.  Returns False if the emission
    was stopped by a handler; otherwise True."""
    
    if self.__fSignalCatchers is None:
      return
    
    return _Emit(self, self.__fSignalCatchers, signal_name, args)

  #----------------------------------------------------------------------
  def emit_stop_by_name(self, name):
    """ Stops single emission of named signal, note that if signal is being
    recursively emitted, only the inner-most one is stopped. """

    _EmitStopByName(self, name)
 
  #----------------------------------------------------------------------
  @classmethod
  def signals(cls):
    """ Returns frozenset of signal names. """

    return _gSignalsForCls.get(cls, _kEmptyFrozenset)

  #----------------------------------------------------------------------
  @classmethod
  def add_signal(cls, signal_name):
    """Add a new signal_name of signal this subclass of destroyable can emit,
    in addition to those set in the constructor"""
    
    signals = cls.signals()
    _gSignalsForCls[cls] = signals.union([signal_name])
    
  #----------------------------------------------------------------------
  def num_catchers(self, signal_name):
    """ Returns the number of signal catchers attached to given signal """

    if self.destroyed():
      return 0
    
    catcher_list = self.__fSignalCatchers.get(signal_name, ())
    return len(catcher_list)
