#########################################################################
"""

Copyright (c) 1999-2012, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel and John P. Ehresman

"""
#########################################################################

import logging
import os
import sys
import socket
import threading
import time


if sys.platform == 'win32':
  from . import win32
else:
  win32 = None
  
if sys.platform != 'win32':
  import fcntl
else:
  fcntl = None
  
try:
  from guiutils import wgtk
except ImportError:
  wgtk = None
  
from debug.tserver import channel

from . import destroyable

class CMainloop(object):
  """ Class to provide mainloop support; serves as an abstract base class.
  There are currently classes that tie in with the wgtk / Qt mainloop
  and that implement a select / win32 wait for based mainloop.  In both
  cases, the timeout_add_while_alive may be called from any thread.
  All callbacks get called on the thread that main is run on.
  """
    
  _gActive = None
        
  def main(self):
    """ Start the main loop.  Must be called on same thread that the instance
    was created on. """

  def mainquit(self):
    """ Quit from the currently running mainloop. """
      
  def timeout_add_while_alive(self, msec, cb, alive_obj):
    """ Add a timeout to call the callback on the thread that main is
    running on.  Use msec=0 to call as soon as possible.  If alive_obj
    is not None, the callback is removed when the alive_obj is 
    destroyed. 
    
    Note that there is no way to remove the timeout except via the 
    alive_obj currently.
    """

  def add_socket_watch(self, sock, cb, readable=False, writable=False):
    """ Add a callback when the socket is readable or writable.
    Returns and object that may be destroyed to remove the watch. """

_gMainloop = None

def SetMainloop(mloop):
  
  global _gMainloop
  
  assert _gMainloop is None or mloop is None  
  
  _gMainloop = mloop
  
def GetMainloop():
  
  if _gMainloop:
    return _gMainloop
  
  wgtk_mainloop = CWgtkMainloop()
  SetMainloop(wgtk_mainloop)
  return wgtk_mainloop
  
  
  


class CWgtkMainloop(object):
    """ Class to provide mainloop support.  Default implementation
    calls down to wgtk but all methode can be overridden for use
    where wgtk isn't available.  Mainloop to be used is gMainloop in
    the module"""
    
    _gRecieverClass = None
    _gCallbackEventClass = None
    
    def __init__(self):
      
        self._init_receiver_classes()
        
        self._fMainThread = threading.current_thread()
        self._fReceiverObject = self._gRecieverClass()
    
    def main(self):
        
        assert wgtk is not None

        
        wgtk.main()

    def mainquit(self):
        
        assert wgtk
        
        wgtk.main_quit()
        
    def timeout_add_while_alive(self, msec, cb, alive_obj):
        
        assert wgtk
        
        def add_timeout():
          wgtk.timeout_add_while_alive(msec, cb, alive_obj)
        
        if threading.current_thread() == self._fMainThread:
          add_timeout()
        else:
          self._run_on_main_thread(add_timeout)
          
    def add_socket_watch(self, sock, cb, readable=False, writable=False):
        
        assert wgtk
        
        flags = ((wgtk.QSocketNotifier.Read if readable else 0)
                 | (wgtk.QSocketNotifier.Write if writable else 0))
        cond = wgtk.QSocketNotifier.Type(flags)
        watcher = wgtk.QSocketNotifier(sock.fileno(), cond)
        watcher.activated.connect(cb)
        return watcher

    def _run_on_main_thread(self, cb):
      
      event = self._gCallbackEventClass(cb)
      wgtk.gQtApp.postEvent(self._fReceiverObject, event)

    @classmethod
    def _init_receiver_classes(cls):
      
      if cls._gRecieverClass is not None:
        return
      
      class CCrossThreadEvent(wgtk.QEvent):
        """ Event subclass to send a schedule a callback to run on the main 
        thread """
              
        def __init__(self, callback):
                
          wgtk.QEvent.__init__(self, wgtk.QEvent.User)
                
          self.fCallback = callback
          
      cls._gCallbackEventClass = CCrossThreadEvent
      
      class CMainThreadReceiverObject(wgtk.QObject):
        """ A receiver object to get callbacks to run on the main thread.  Works
        by having CCrossThreadEvent objects posted to an object created on the
        main thread (it can be a global).  The fCallback is simply called from
        the event method. """
      
        def event(self, ev):
          
          if isinstance(ev, CWgtkMainloop._gCallbackEventClass):
            ev.fCallback()
            
          return True

      cls._gRecieverClass = CMainThreadReceiverObject
      
class CNonblockingSocketMainloop(object):
  """ A mainloop implementation that uses select or the win32 WaitForObjects
  to wait for socket activity.  All sockets watched are set to nonblocking 
  because of win32 limitation.  All methods are thread safe after __init__()
  completes. """
  
  # Whether to use the win32 WaitFor... family of functions and supporting api's.
  # If false on win32, uses a socket pair as created by remoteops.SocketPair
  _kUseWin32WaitFor = (sys.platform == 'win32')
  
  class _CTimeout:
    """ Data class for timeout info """
    
    def __init__(self, msec, cb):
      
      self.cb = cb
      self.orig_msec = msec
      self.msec_left = msec
      
  class _CSocketWatcher(destroyable.CDestroyable):
    """ Info for a single socket being watched.  It is a destroyable so
    the watch may be removed with a .destroy() call. """
    
    def __init__(self, sock, cb, readable, writable, mainloop):
      
      destroyable.CDestroyable.__init__(self)
      
      self.sock = sock
      self.cb = cb
      self.readable = readable
      self.writable = writable
      self.mainloop = mainloop
      
    def _destroy_impl(self):
      
      self.mainloop.remove_socket_watch(self.sock)
      
  class _CSingleSelectWait:
    """ Data class for a single wait call when using select """
    
    def __init__(self):
      
      self.rsockets = []
      self.wsockets = []
      self.esockets = []
      self.timeout = None
      
  class _CSingleWaitForWait:
    """ Data class for a single wait call when using win32 WaitFor... functions """
    
    def __init__(self):
      
      self.socket_pairs = []
      self.timeout = None
      
  def __init__(self):
    
    self._fMainCalled = False
  
    self._fQuitCalled = False
    self._fSocketsWatched = {}
    self._fSelfLock = threading.RLock()
    self._fTimeouts = []
    
    self._fInterruptPipe = None
    self._fInterruptSocketPair = None
    self._fInterruptWin32Event = None
    
    self._fEventsForSockets = {}
    
    if self._kUseWin32WaitFor:
      self._fInterruptWin32Event = win32.CreateEvent(None, True, False, None)
    elif sys.platform != 'win32':
      self._fInterruptPipe = os.pipe()
      for end in self._fInterruptPipe:
        fcntl.fcntl(end, fcntl.F_SETFL, os.O_NONBLOCK)
    else: # win32
      from . import remoteops
      self._fInterruptSocketPair = remoteops.SocketPair()
      for end in self._fInterruptSocketPair:
        end.setblocking(False)
        
  def main(self):
    
    self._fMainCalled = True
    self._fLastTimeCheck = time.time()
    
    while not self._fQuitCalled:

      single_watch = self._prepare_single_wait()
      
      if self._kUseWin32WaitFor:
        available = self._wait_with_win32_wait_for(single_watch)
      else:
        available = self._wait_with_select(single_watch)
      
      self._call_callbacks(available)
      
    self._cleanup()

  def _prepare_single_wait(self):
    """ Prepare for a single wait.  Returns a _CSingleWait instance. """
    
    with self._fSelfLock:
      if self._kUseWin32WaitFor:
        single_watch = self._CSingleWaitForWait()
  
        for sock in self._fSocketsWatched.keys():
          event = self._fEventsForSockets.get(sock)
          if event is not None:
            single_watch.socket_pairs.append((sock, event))

      else:
        single_watch = self._CSingleSelectWait()
  
        for sock, watch in self._fSocketsWatched.items():
          if watch.readable:
            single_watch.rsockets.append(sock)
          if watch.writable:
            single_watch.rsockets.append(sock)
          single_watch.esockets.append(sock)
        
      self._update_timeout_msec()
      if len(self._fTimeouts) == 0:
        timeout_secs = None
      else:
        min_msec = min(timeout.msec_left for timeout in self._fTimeouts)
        single_watch.timeout = max(0, min_msec / 1000.0)
    
    return single_watch    
          
  def _wait_with_select(self, single_watch):
    """ Wait for socket activity or timeout to expire using a select call.
    Must be called when the _fSelfLock is not held. """
        
    # _fSelfLock is never held
    assert not self._fSelfLock._is_owned()

    if self._fInterruptPipe is not None:
      rd_end = self._fInterruptPipe[0]
    else:
      assert self._fInterruptSocketPair is not None
      rd_end = self._fInterruptSocketPair[0]
    
    rsockets = list(single_watch.rsockets)
    rsockets.append(rd_end)
    
    reads, writes, expts = channel.safe_select(rsockets, single_watch.wsockets, 
                                               single_watch.esockets, single_watch.timeout)
    
    all_socks = set(reads + writes + expts)
    
    # If interrupt pipe was written to, read from it before returning and
    # acquiring _fSelfLock so that if a write happens between the select
    # return and now, it's read from before acquiring _fSelfLock.  Any change
    # will be handled within the _fSelfLock'd block.  If a write happens
    # after the read, it may or may not be handled in the _fSelfLock'd 
    # block but it will interrupt the select the next time around
    if rd_end in all_socks:
      all_socks.remove(rd_end)
      if isinstance(rd_end, int):
        d = os.read(rd_end, 1000) # Data doesn't matter; pipe is non-blocking
      else:
        d = rd_end.recv(1000)
    
    return all_socks
    
  def _wait_with_win32_wait_for(self, single_watch):
    """ Wait for socket activity or timeout to expire using a select call.
    Must be called when the _fSelfLock is not held. """

    assert self._kUseWin32WaitFor
    
    handles = [event for sock, event in single_watch.socket_pairs]
    handles.append(self._fInterruptWin32Event)
    
    msec_timeout = (win32.INFINITE if single_watch.timeout is None
                    else int(single_watch.timeout * 1000))
    
    r = win32.WaitForMultipleObjectsEx(handles, False, msec_timeout, True)

    win32.ResetEvent(self._fInterruptWin32Event)

    available = set()
    
    for sock, event in single_watch.socket_pairs:
      if win32.WaitForSingleObjectEx(event, 0, False) == win32.WAIT_OBJECT_0:
        available.add(sock)
        win32.ResetEvent(event)
        
    return available
    
  def _call_callbacks(self, available):
    """ Call all callbacks after the wait for single_wait has been done.
    The _fSelfLock should not be held when this is called. """
    
    with self._fSelfLock:
      watch_iter = (self._fSocketsWatched.get(sock) for sock in available)
      sock_cbs_to_call = [watch.cb for watch in watch_iter 
                          if (watch is not None and not watch.destroyed()
                              and watch.cb is not None)]

      self._update_timeout_msec()
      timeouts_to_call = [t for t in self._fTimeouts if t.msec_left <= 0]

    # Make callbacks after releasing _fSelfLock
    for cb in sock_cbs_to_call:
      try:
        cb()
      except:
        logging.exception("Exception during socket callback: %r" % (cb,))
          
    for timeout in timeouts_to_call:
      if timeout.cb is None:
        remove = True
      else:
        try:
          ret = timeout.cb()
        except:
          logging.exception("Exception during timeout: %r" % (timeout.cb,))
          remove = False
        else:
          remove = not ret
        
      if remove:
        self.timeout_remove(timeout)
      else:
        timeout.msec_left = timeout.orig_msec
          
  def _update_timeout_msec(self):
    """ Update all timeouts, reducing the msec for each by the time since 
    the last time this method was calle.  Must be called with _fSelfLock
    held. """
    
    assert self._fSelfLock._is_owned()
    
    now = time.time()
    msec_diff = max(0, int((now - self._fLastTimeCheck) * 1000))
    self._fLastTimeCheck = now
    
    for timeout in self._fTimeouts:
      timeout.msec_left -= msec_diff 

  def _interrupt_wait(self):
    """ Interrupt any current wait. """
    
    if self._fInterruptWin32Event is not None:
      win32.SetEvent(self._fInterruptWin32Event)
    elif self._fInterruptSocketPair is not None:
      self._fInterruptSocketPair[1].send('1')
    else:
      os.write(self._fInterruptPipe[1], '1')
  
  def mainquit(self):
  
    self._fQuitCalled = True
    self._interrupt_wait()
      
  def timeout_remove(self, timeout):
    
    with self._fSelfLock:
      try:
        self._fTimeouts.remove(timeout)
      except ValueError:
        pass
      else:
        self._interrupt_wait()
        
  def timeout_add_while_alive(self, msec, cb, alive_obj):
    
    def on_destroy(obj, timeout):
      self.timeout_remove(timeout)
    
    with self._fSelfLock:
      timeout = self._CTimeout(msec, cb)
      if alive_obj is not None:
        alive_obj.connect('destroy', on_destroy, timeout)
      self._fTimeouts.append(timeout)
      
    self._interrupt_wait()
    return timeout
  
  def add_socket_watch(self, sock, cb, readable=False, writable=False):
      
    sock.setblocking(False)
    
    with self._fSelfLock:
      watch = self._CSocketWatcher(sock, cb, readable, writable, self)
      self._fSocketsWatched[sock] = watch

      if self._kUseWin32WaitFor:
        event = self._fEventsForSockets.get(sock)
        if event is None:
          event = win32.CreateEvent(None, True, False, None)
          self._fEventsForSockets[sock] = event
          
        flags = 0
        if readable:
          flags |= win32.FD_READ | win32.FD_CLOSE
        if writable:
          flags |= win32.FD_WRITE
        win32.WSAEventSelect(sock, event, flags)

    self._interrupt_wait()
    return watch
  
  def remove_socket_watch(self, sock):
    
    with self._fSelfLock:
      print 'pop'
      self._fSocketsWatched.pop(sock, None)

      if win32 is not None:
        event = self._fEventsForSockets.pop(sock, None)
        if event is not None:
          try:
            win32.WSAEventSelect(sock, event, 0)
          except socket.error:
            pass # Socket is probably closed
          win32.CloseHandle(event)
        
  def _cleanup(self):
    
    if self._fInterruptWin32Event is not None:
      win32.CloseHandle(self._fInterruptWin32Event)
      self._fInterruptWin32Event = None
      
    for sock, event in self._fEventsForSockets.items():
      try:
        win32.WSAEventSelect(sock, event, 0)
      except socket.error:
        pass # Socket is probably closed
      win32.CloseHandle(event)
    self._fEventsForSockets.clear()