""" reflect.py -- Reflection for Python objects

Copyright (c) 1999-2012, Archaeopteryx Software, Inc.  All rights reserved.

Written by Stephan R.A. Deibel and John P. Ehresman

"""
#########################################################################

import collections
import gc
import os
import types
import sys
import time
import string
import traceback
  
# Bitwise flag values used in co_flags 
kVarListArg = 1 << 2
kKeywordDictArg = 1 << 3

# List of file objects and/or file names to write exception info to
kExceptionLogFiles = []
kExceptionTruncateThreshold = None
kExceptionLogConfigured = 0
kExceptionGUICallback = None

#-----------------------------------------------------------------------
def GetArgCount(clb):
  """ Utility to count number of args for given callable object """

  assert callable(clb), "Not a callable object"

  if type(clb) is types.FunctionType: # aka LambdaType
    return clb.func_code.co_argcount
  elif type(clb) is types.MethodType:
    return clb.im_func.func_code.co_argcount
  elif type(clb) is types.ClassType:
    return clb.__init__.im_func.func_code.co_argcount
  elif type(clb) is types.InstanceType and hasattr(clb, '__call__'):
    return clb.__call__.im_func.func_code.co_argcount
  else:
    assert 0, "Unhandled callable type"


#-----------------------------------------------------------------------
class CArgDescription:
  """ Value object to contain description of single argument. """

  #-----------------------------------------------------------------------
  def __init__(self, name, has_default = 0, default_value = None, is_varlist = 0,
               is_kwdict = 0):
    """ Constructor; arguments set attributes which are invarient. """
    
    self.name = name
    self.has_default = has_default
    self.default_value = default_value
    self.is_varlist = is_varlist
    self.is_kwdict = is_kwdict

#-----------------------------------------------------------------------
class CAttribDescription:
  """ Value object to contain description of attribute found in
  a class or a class instance. """

  #-----------------------------------------------------------------------
  def __init__(self, name, value, found_in):
    """ Constructor; arguments set attributes which are invarient.  found_in
    is a reference to the instance or class the attribute was found in. """

    assert isinstance(found_in, (types.InstanceType, types.ClassType, type))
    
    self.name = name
    self.value = value
    self.found_in = found_in

#----------------------------------------------------------------------
class CLoggedOutput:
  """ Simple class to redirect output to stdout/stderr """
  
  def __init__(self):
    self.fCurLine = ""
    pass
    
  def write(self, str):
    try:
      try:
        self.fCurLine = self.fCurLine + str
      except UnicodeDecodeError:
        fs_encoding = sys.getfilesystemencoding()
        if fs_encoding is None:
          fs_encoding = sys.getdefaultencoding()
        try:
          self.fCurLine = self.fCurLine + unicode(str, fs_encoding)
        except UnicodeDecodeError:
          self.fCurLine = self.fCurLine + repr(str)
      if '\n' in self.fCurLine and kExceptionLogConfigured:
        lines = self.fCurLine.split('\n')
        self.fCurLine = lines[-1]
        old_lines = '\n'.join(lines[:-1])
        ReportMessageString(old_lines)
    except:
      global kExceptionLogFiles
      save = kExceptionLogFiles
      try:
        sys.__stderr__.write(('#' * 80) + '\n')
        sys.__stderr__.write('An exception occurred when writing the exception log\n')
        sys.__stderr__.write('Logging to sys.__stderr__ instead\n')
        kExceptionLogFiles = [sys.__stderr__]
        ReportCurrentException(suppress_gui=True)
        sys.__stderr__.write('Output str was:\n') 
        try:
          sys.__stderr__.write(repr(str))
        except:
          sys.__stderr__.write("[FAILURE: could not write repr(str) either]\n")
      except:
        sys.__stderr__.write('Sorry, could not write traceback either (may be recursion depth error)\n')
      kExceptionLogFiles = save
      
  # Added so modules that expect it are happy
  def flush(self):
    pass
  

#-----------------------------------------------------------------------
def GetArgDescriptionList(clb):
  """ Get list of argument description objects.  Does not return the entry for
  self for a method instance or a class initializer. """
  
  assert callable(clb), "Not a callable object"

  if type(clb) is types.FunctionType: # aka LambdaType
    func = clb
    first_arg = 0
  elif type(clb) is types.MethodType:
    func = clb.im_func
    first_arg = 1
  elif type(clb) is types.ClassType:
    func = clb.__init__.im_func
    first_arg = 1
  elif type(clb) is types.InstanceType and hasattr(clb, '__call__'):
    func = clb.__call__.im_func
    first_arg = 0
  else:
    assert 0, "Unhandled callable type"

  # Process positional args
  arg_list = []
  count = func.func_code.co_argcount
  if func.func_defaults is None:
    first_default = count
  else:
    first_default = count - len(func.func_defaults)
  for i in range(first_arg, count):
    name = func.func_code.co_varnames[i]
    if i < first_default:
      arg = CArgDescription(name)
    else:
      arg = CArgDescription(name, 1, func.func_defaults[i - first_default])
    arg_list.append(arg)

  # Check for varlist
  if func.func_code.co_flags & kVarListArg:
    arg = CArgDescription(func.func_code.co_varnames[count], is_varlist = 1)
    count = count + 1
    arg_list.append(arg)

  # Check for kw arg
  if func.func_code.co_flags & kKeywordDictArg:
    arg = CArgDescription(func.func_code.co_varnames[count], is_kwdict = 1)
    count = count + 1
    arg_list.append(arg)

  return arg_list 

#-----------------------------------------------------------------------
def GetKnownAttribsOfClass(cl, include_overridden):
  """ Finds the known attributes of a class and returns them as a list
  of CAttribDescription's.  A known attribute is one that is recorded
  in the classes's dictionary or in it's base classes. """

  assert isinstance(cl, (types.ClassType, type))

  attrs = []
  unique_names = {}
  for name, value in cl.__dict__.items():
    attrs.append(CAttribDescription(name, value, cl))
    unique_names[name] = 1
    
  for base in cl.__bases__:
    base_attr_list = GetKnownAttribsOfClass(base, include_overridden)
    for base_attr in base_attr_list:
      name = base_attr.name
      if include_overridden or not unique_names.has_key(name):
        attrs.append(base_attr)
        unique_names[name] = 1
        
  return attrs

#-----------------------------------------------------------------------
def GetKnownAttribsOfInstance(inst):
  """ Finds the known attributes of an instance and returns them as a
  list of CAttribDescription's.  A known attribute is one that is
  recorded in the instance's dictionary or in it's class. """

  assert isinstance(inst, types.InstanceType)

  attrs = []
  unique_names = {}
  for name, value in inst.__dict__.items():
    attrs.append(CAttribDescription(name, value, inst))
    unique_names[name] = 1
    
  cl_attrs = GetKnownAttribsOfClass(inst.__class__, 0)
  for cl_attr in cl_attrs:
    if not unique_names.has_key(cl_attr.name):
      attrs.append(cl_attr)
  return attrs

#-----------------------------------------------------------------------
def GetCurrentFrame():
  """ Returns frame object of the calling function. """

  try:
    raise ZeroDivisionError
  except ZeroDivisionError:
    f = sys.exc_info()[2].tb_frame.f_back
    return f

#-----------------------------------------------------------------------
def ConfigureExceptionReporting(output_files, truncate_at=100000):
  """ Set the file(s) where ReportCurrentException places output.
  Pass in a list of file objects and strings.  Strings are interpreted
  as file names.  Set truncate_at to max byte size of the log file. """
  
  global kExceptionLogFiles
  global kExceptionTruncateThreshold
  global kExceptionLogConfigured
  
  kExceptionLogFiles = output_files
  kExceptionTruncateThreshold = truncate_at
  kExceptionLogConfigured = 1
  
#-----------------------------------------------------------------------
def GetCurrentException(exc_info=None, frames_above=()):
  """ Get the current exception as list of strings.  Uses given
  exc_info if not None, or otherwise sys.exc_info(). """

  # Build exception message lines
  if exc_info is not None:
    t, val, tb = exc_info
  else:
    t, val, tb = sys.exc_info()

  exc = []
  try:
    exc.append("Runtime failure details:")
    exc.append("Exception: %s" % repr(str(t)))
    exc.append("Value = %s" % repr(str(val)))
    exc.append("Traceback:")
    tbf = traceback.format_tb(tb)
    exc.extend(frames_above)
    for item in tbf:
      exc.append(unicode(item).rstrip())
    if t is SyntaxError:
      try:
        exc.append('  File "%s", line %i:\n    %s\n' % (val[1][0], val[1][1], val[1][3]))
      except:
        pass
  except:
    pass
  
  return exc

#-----------------------------------------------------------------------
def GetCurrentExceptionString(exc_info=None):
  """ Get the current exception as multi-line string.  Uses
  given exc_info if not None or otherwise sys.exc_info() """

  return string.join(GetCurrentException(exc_info), '\n')

#-----------------------------------------------------------------------
def ReportCurrentException(suppress_gui=False, exc_info=None, msg=None,
                           frames_above=None):
  """ Reports the current exception. """

  if exc_info is None:
    exc_info = sys.exc_info()

  # Report in wing's debugger if not suppresssing gui
  if 'WINGDB_ACTIVE' in os.environ:
    sys.excepthook(*exc_info)
    
  # Write to each file and truncate log files
  if frames_above is None:
    frames_above = traceback.format_stack()
    frames_above = frames_above[:-1]
  exc = GetCurrentException(exc_info, frames_above)
  __WriteExceptionLogEntries(exc)

  # Also call hook if present
  if not suppress_gui and callable(kExceptionGUICallback):
    exc = GetCurrentException(exc_info)
    kExceptionGUICallback(exc)

#-----------------------------------------------------------------------
def ReportMessageString(s):
  """ Write a message to the exception log """
  
  # Build message lines
  s = string.replace(s, os.linesep, '\n')
  
  # Write to each file and truncate log files
  __WriteExceptionLogEntries([s,])
  
#-----------------------------------------------------------------------
gLastLines = {}
def __WriteExceptionLogEntries(entry):
  """ Write entry in the configured exception log files """
  
  # Build time stamp
  dtime = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(time.time()))
  
  # Write entries
  for f in kExceptionLogFiles:
    fo = None
    if isinstance(f, basestring):
      try:
        try:
          fo = open(f, 'a')
        except IOError:
          if not os.path.exists(os.path.dirname(f)):
            os.makedirs(os.path.dirname(f))
            fo = open(f, 'a')
      except:
        fo = None
    else:
      fo = f

    if fo != None:
      if not f in gLastLines:
        gLastLines[f] = ['', 0]
      try:
        try:          
          for line in entry:
            sublines = line.split('\n')  # Yes, it's \n on win32 also
            for sline in sublines:
              if sline == gLastLines[f][0]:
                gLastLines[f][1] += 1
              else:
                if gLastLines[f][1] > 0:
                  fo.write("wingide - %s - (repeated %i times)\n" % (dtime, gLastLines[f][1]))
                gLastLines[f] = [sline, 0]
                if sline.strip():
                  if not sline.startswith('wingdbg - 2'):
                    fo.write("wingide - %s - %s\n" % (dtime, sline))
                  else:
                    fo.write(sline + '\n')
        finally:    
          if isinstance(f, basestring):
            fo.close()
            __TruncateExceptionFile(f)
          else:
            fo.flush()
      except: # Keep going if there are any exceptions
        pass

#-----------------------------------------------------------------------
def __TruncateExceptionFile(f):
  """ Truncate an exception log file kept by ReportCurrentException() """
  
  import stat

  fsize = os.stat(f)[stat.ST_SIZE]
  
  if kExceptionTruncateThreshold is not None and fsize > kExceptionTruncateThreshold:
    fo = open(f)
    try:
      txt = fo.read()
    finally:
      fo.close()
    txt = txt[len(txt) - int(0.80 * kExceptionTruncateThreshold):]
    nlpos = string.find(txt, '\n')
    txt = txt[nlpos+1:]
    fo = open(f, 'wb')
    try:
      fo.write(txt)
    finally:
      fo.close()

#-----------------------------------------------------------------------
def DbgGetAllObjectsByType():
  """ Returns dictionary of all objects in the process, keyed by type
  of object.  Only works in pydebug builds, otherwise returns None.
  """

  try:
    getobjects = sys.getobjects
  except AttributeError:
    return None
  
  bytype = collections.defaultdict(list)
  for obj in getobjects(0):
    bytype[type(obj)].append(obj)
  return bytype

#-----------------------------------------------------------------------
def DbgPrintAllObjectsTypeCounts(bytype=None):
  """ Prints count of all objects of each type in a pydebug build. """
  
  if bytype is None:
    bytype = DbgGetAllObjectsByType()
    if bytype is None:
      return
  
  pairs = sorted((len(obj_list), repr(obj_type))
                 for obj_type, obj_list in bytype.iteritems())
  for count, obj_type in pairs:
    print '%s instances: %s' % (count, repr(obj_type))

class CHeapSnapshot(object):
  """ Take a snapshot of the heap or certain objects in it for debugging
  purposes.  Uses the gc module; could be sped up via C code. """
  
  def __init__(self, object_seq=None, find_ancestors_count=0):
    """ If object_seq is specified, the objects and their children are scanned
    for their relationship. """

    self._fAllObjects = {}
    """ Dict with all objects found, mapping id(obj) -> obj """
    self._fChildren = collections.defaultdict(set)
    """ Dict of all id's of children, keyed by id of the parent """
    self._fParents = collections.defaultdict(set)
    """ Dict of all id's of parents, keyed by id of the child """

    if object_seq is None:
      scan_all = True
      object_seq = gc.get_objects()
    else:
      scan_all = False
      
    local_frame = sys._getframe()
    scan_up_queue = []
    next_scan_up_queue = []
    ignore_id_set = set([id(local_frame), id(self._fAllObjects), id(object_seq),
                         id(scan_up_queue), id(next_scan_up_queue)])

    for i in xrange(len(object_seq)):
      parent = object_seq[i]
      parent_id = id(parent)
      self._fAllObjects[parent_id] = parent
      children = gc.get_referents(parent)
      for child in children:
        child_id = id(child)
        
        self._fAllObjects[child_id] = child
        self._fChildren[parent_id].add(child_id)
        self._fParents[child_id].add(parent_id)

      if not scan_all and (isinstance(parent, dict) or find_ancestors_count > 0):
        scan_up_queue.append(parent)
        
    found_parents = set()
    i = 0
    max_up_levels = 2
    while i < max_up_levels:
      i += 1
      
      while len(scan_up_queue) > 0:
        obj = scan_up_queue.pop()
        obj_id = id(obj)
        if obj_id not in found_parents:
          found_parents.add(obj_id)
          grand_id_list = self._FindAdditionalParents(obj, id(object_seq))
          for grand_id, grand in grand_id_list:
            if grand_id not in ignore_id_set:
              self._fParents[obj_id].add(grand_id)
              self._fAllObjects[grand_id] = grand
              
              if i + 1 < find_ancestors_count:
                next_scan_up_queue.append(grand)

      scan_up_queue.extend(next_scan_up_queue)
      del next_scan_up_queue[:]
      
  def Clear(self):
    """ Clear all info and release references. """
    
    self._fAllObjects.clear()
    self._fParents.clear()
    self._fChildren.clear()

  def _FindAdditionalParents(self, obj, ignore_ids):
    """ Find parents of obj with the gc module.  Used for scanning up.
    Returns list of (object-id, object) pairs """
    
    local_frame = sys._getframe()
    
    parents = gc.get_referrers(obj)
    pair_list = [(id(p), p) for p in parents if p is not local_frame]
  
    return pair_list
  
  def Discard(self, obj):
    """ Discard knowledge of given object.  Useful for removing containers
    used to collect objects to look at from data structures. """
    
    obj_id = id(obj)
    child_id_iter = self._fChildren.pop(obj_id, ())
    for child_id in child_id_iter:
      parent_id_set = self._fParents.get(child_id)
      if parent_id_set is not None:
        parent_id_set.discard(obj_id)

    parent_id_iter = self._fParents.pop(obj_id, ())
    for parent_id in parent_id_iter:
      child_id_set = self._fChildren.get(parent_id)
      if child_id_set is not None:
        child_id_set.discard(obj_id)
        
    self._fAllObjects.pop(obj_id, None)
  
  def GetParents(self, obj, skip_impl_dicts=True):
    """ Get the known parents of the object.  If skip_impl_dict, then
    any dict instance that is a __dict__ for another object will be
    replaced by the parent of the dict. """
    
    parent_id_list = self._fParents.get(id(obj))
    if parent_id_list is None:
      return None
    
    parent_list = []
    for parent_id in parent_id_list:
      parent = self._fAllObjects.get(parent_id)
      if parent is not None:
        if skip_impl_dicts and isinstance(parent, dict):
          impl_dict_parent = self._LookupKnownImplDictParent(parent, parent_id)
          if impl_dict_parent is not None:
            parent = impl_dict_parent
        parent_list.append(parent)
        
    return parent_list

  def _LookupKnownImplDictParent(self, obj, obj_id):
    """ Find the parent of an impl dict instance if it's known. """
    
    parent_id_seq = self._fParents.get(obj_id, ())
    for parent_id in parent_id_seq:
      parent = self._fAllObjects.get(parent_id)
      if parent is not None:
        try:
          dict_attr = parent.__dict__
        except AttributeError:
          pass
        else:
          if obj is dict_attr:
            return parent
          
    return None
