#########################################################################
""" remoteops.py -- simple rpc for executing operations in remote processes

Copyright (c) 1999-2012, Archaeopteryx Software, Inc.  All rights reserved.

Written by John P. Ehresman and Stephan R.A. Deibel

The basic idea is to expose functionality via objects with asyncronous
methods.  The methods are async because they will be executed after the
calling invocation returns and then deliver the return value or exception
via a callback,  The callback for the return value is specified via 
a callback= keyword argument pass into the method invocation that initiates
the async method; the callback for the exception can be likewise be 
specified via an exception= keyword argument or via a default exception
callback.

For example, a simple echo method would be invoked via:

    def string_received(s):
        ....
    obj.echo("Hello World", callback=string_received)
    
The obj.echo call will return immediately and the string_received function
will be called when the result is available.  The implementation of the
echo method might look like:

    def echo(s):
        return s
        
Objects are exposed via CContext objects.  Each context is connected to
another context; the context that code works with is the local context and
the other context is the remote context -- the two are symmetrical; the 
remote context is local to the process / code using it and treats its
twin as the remote context.  Objects are made available to the other context
by registering them as network accessible objects; only these objects are
exposed to the other channel.  Each context can have a primary network
accessible object that the other context can access without any other 
setup.

Proxy objects are CDestroyable objects which can be destroyed.  If a proxy
is destroyed, all callbacks for method calls initiated but have not been
called (because the result hasn't arrived) will _not_ be called unless
a the call was initiated with a callback_if_destroyed=True keyword argument
is specified.

"""

import contextlib
import datetime
import exceptions
import errno
import cStringIO
import cPickle
import logging
import os

import socket

import sys
import time

from debug.tserver import channel
try:
    from guiutils import wgtk
except ImportError:
    wgtk = None
    
from . import destroyable
from . import mainloop

try:
    from . import spawn
except ImportError:
    spawn = None

kShutdown = 0x000001

kNamedMethodCall = 0x00101
kNoResultNamedMethodCall = 0x00102
kMethodCallResult = 0x00103

kLocalHost = '127.0.0.1'

def Destroy(obj):
    
    if wgtk:
        wgtk.Destroy(obj)
    else:
        obj.destroy()

def GetMainloop():
    
    return mainloop.GetMainloop()

def SocketPair(family=None, type=socket.SOCK_STREAM, proto=0):
    """ Creats a connected pair of sockets.  Works on win32 by creating
    a listening soket on localhost and then connecting to it.  The 
    family arg defaults to AF_INET on win32 and AF_UNIX elsewhere. """
    
    if sys.platform != 'win32':
        if family is None:
            family = socket.AF_UNIX
        return socket.socketpair(family, type, proto)

    if family is None:
        family = socket.AF_INET
        
    accepted = None
    conn = None
    listener = socket.socket(family=family, type=type, proto=proto)
    try:
        listener.bind(('127.0.0.1', 0))
        listener.listen(16)
        conn = socket.socket(family=family, type=type, proto=proto)
        conn.connect(listener.getsockname())
        while True:
            accepted, addr = listener.accept()
            if addr == conn.getsockname():
                return accepted, conn
    except Exception:
        if accepted:
            accepted.close()
        if conn:
            conn.close()
        raise
    finally:
        listener.close()
        
def GetContext(obj):
    """ Return the context the object is registered with.  None is returned
    if object is not registed with any context. """
    
    for ctx in CContext._gAllContexts:
        obj_id = ctx.fSerializer.GetObjectID(obj)
        if ctx.fSerializer.GetLocalObject(obj_id) == obj:
            return ctx
        
    return None

class CSerializer(object):
    """ Serializes data to go over the network using pickle.  Maintains a
    dict of objects that can be accessed over the net and transmits references
    to them.
    
    Additional types may be allowed to be transferred; the default pickle
    mechanisms are used to serialize these.  Any other types are not
    allowed and an exception will be raised either in Serialize() or 
    Deserialize()
    """
    
    kPlainTypes = (int, float, str, unicode, dict, tuple, list, set, bool)

    def __init__(self, context):
        
        self.fContext = context
        
        self._fAllowedTypesByNamePair = {}
        self._fAllowedTypes = set()
        
        self._fNetworkAccessibleObjects = {}
        self._fPersistentIDs = {}
        self._fOverriddenPersistantIDs = {}
        
        self._fActiveProxies = {}
        
        self.AddAllowedTypes(self.kPlainTypes)
        
    def AddAllowedTypes(self, allowed_types):
        """ Add sequence of types to be allowed. """
        
        for allowed in allowed_types:
            self._fAllowedTypes.add(allowed)
            pair = (allowed.__module__, allowed.__name__)
            self._fAllowedTypesByNamePair[pair] = allowed

    def AddNetworkAccessibleObject(self, obj, obj_id=None):
        """ Add object that can be exposed over the network. """
        
        obj_id = self.GetObjectID(obj)
        self._fNetworkAccessibleObjects[obj_id] = obj
        self._fPersistentIDs[obj] = obj_id
        return obj_id
        
    def RemoveNetworkAccessibleObject(self, obj, obj_id=None):
        """ Removes object previously added. """
        
        obj_id = self.GetObjectID(obj)
        self._fNetworkAccessibleObjects.pop(obj_id, None)
        self._fPersistentIDs.pop(obj, None)
        return obj_id

    def GetObjectID(self, obj):
        """ Return object id; unique within the process. """
        
        if obj == self.fContext.fPrimaryObject:
            return hex(0)
        
        return hex(id(obj))
    
    def GetProxy(self, obj_id):
        """ Returns a proxy for given object in the other process. """
        
        proxy = self._fActiveProxies.get(obj_id)
        if proxy is None:
            proxy = CProxy(self.fContext, obj_id)
            self._fActiveProxies[obj_id] = proxy
            
        return proxy
    
    def RemoveActiveProxy(self, proxy):
        """ Removes from list of active proxies """
        
        self._fActiveProxies.pop(proxy._fObjID, None)
    
    def GetLocalObject(self, local_id):
        """ Get the local object identified by local_id """
        
        return self._fNetworkAccessibleObjects.get(local_id)
    
    def Serialize(self, py_data):
        """ Serialize data and return byte string. """
        
        buff = cStringIO.StringIO()
        p = cPickle.Pickler(buff)
        p.persistent_id = self._PersistentID
        p.dump(py_data)
        return buff.getvalue()

    def Deserialize(self, bytes_data):
        """ Deserialize byte string and return python objects """
        
        buff = cStringIO.StringIO(bytes_data)
        p = cPickle.Unpickler(buff)
        p.persistent_load = self._PersistentLoad
        def find_global(modname, name):
            return self._fAllowedTypesByNamePair.get((modname, name))
            
        p.find_global = find_global
        return p.load()
        
    def _PersistentID(self, obj):
        """ Method used by pickle to return a byte string for object.  Called
        for each and every object pickled (incuding subobjects) """
        
        # Return None to let the default pickle machinery to work; note that 
        # the type object for any allowed type must be allowed and not only
        # the instances of the type
        if obj is None or type(obj) in self._fAllowedTypes or obj in self._fAllowedTypes:
            return None
        
        obj_id = self.GetObjectID(obj)
        if obj_id in self._fNetworkAccessibleObjects:
            otype = 'local'
        elif isinstance(obj, CProxy):
            otype = 'proxy'
            obj_id = obj._fObjID
        else:
            raise ValueError('Cannot send object over network: %r' % (obj,))
        
        persist_id = 'net:%s:%s' % (otype, obj_id)
        return persist_id
         
    def _PersistentLoad(self, persist_id):
        """ Method used by pickle to return object for byte string returned
        by _PersistentID.  Only called on byte string returned by _PersistentID
        """

        parts = persist_id.split(':')
        prefix = parts[0]
        if prefix == 'net':
            if len(parts) != 3:
                raise ValueError('Unable to load object for persistent id: %r'
                                 % (persist_id))
            
            otype = parts[1]
            obj_id = parts[2]
            if otype == 'local':
                return self.GetProxy(obj_id)
            elif otype == 'proxy':
                return self._fNetworkAccessibleObjects[obj_id]
        
        raise ValueError('Unable to load object for persistent id: %r'
                         % (persist_id))

        

class CContext(destroyable.CDestroyable):
    """ Context for one side of remote communications.  Knows all objects
    that can be accessed over the network and how to serialize references
    to them.  When objects are deserialized, proxies are created for remote
    objects. 
    
    Handles exceptions by sending the exception value and the list of strings
    returned by traceback.extract_tb over the wire.  The exception object
    is passed to the exception callback if one is available after the extracted
    list of strings are attached to it as the object's _remote_traceback_extract
    attribute.  
    """
    
    
    kSignals = (
        # Emitted when shutdown is called or a shutdown message is received;
        # it is only sent once
        'shutdown', 
        # Emitted when a connection is connected to its remote half
        'connected',
    )
    
    kLogActivity = False
    
    _gAllContexts = set()

    def __init__(self):

        destroyable.CDestroyable.__init__(self, self.kSignals)
        
        self._fOtherContextDirect = None
        
        self.fPrimaryObject = None
        self.fSerializer = CSerializer(self)

        all_exc_types = (exc_type for exc_type in exceptions.__dict__.values()
                         if type(exc_type) == type and issubclass(exc_type, Exception))
        self.fSerializer.AddAllowedTypes(all_exc_types)
        self.fSerializer.AddAllowedTypes([datetime.datetime])
        
        self._fNetChannel = None
        self._fWaitingForCommand = 0
        
        self._fNextActiveRequestID = 0
        self._fCallbacksForActiveRequests = {}
        self._fDefaultExceptionCallback = None
        
        self._fDiscardCallsObjectIDs = set()
        
        self._fSocketWaiter = None

        self._fShutdownSent = False
        
        # List of calls initiated before net channel is set
        self._fPendingOutgoingCalls = []
        
        self._gAllContexts.add(self)
        
    def _destroy_impl(self):
        
        for proxy in list(self.fSerializer._fActiveProxies.values()):
            proxy.destroy()
            
        self.Shutdown(immediate=True)
        self._gAllContexts.discard(self)

    def _LogActivity(self, msg, *args):

        if not self.kLogActivity:
            return
        
        if len(args):
            s = msg % args
        else:
            s = msg
            
        print(time.strftime('%M:%S') + ' -- ' + s)
        sys.stdout.flush()
        
    def AddNetworkAccessibleObject(self, obj, obj_id=None):
        """ Add object that can be exposed over the network. """
    
        assert obj is not None

        obj_id = self.fSerializer.AddNetworkAccessibleObject(obj, obj_id)
        self._fDiscardCallsObjectIDs.discard(obj_id)
        
    def RemoveNetworkAccessibleObject(self, obj, obj_id=None, discard_calls=False):
        """ Removes object previously added. """
        
        assert obj is not None
        
        obj_id = self.fSerializer.RemoveNetworkAccessibleObject(obj, obj_id)
        
        if discard_calls:
            self._fDiscardCallsObjectIDs.add(obj_id)
                
        
    def Shutdown(self, immediate=False):
        """ Shuts down any connection with other process """
        
        if not self._fShutdownSent:
            self._fShutdownSent = True
            if self._fNetChannel:
                self._SendDataTuple((kShutdown,), ignore_exceptions=True)
            self.emit('shutdown', self)
        
        if self._fOtherContextDirect:
            self._fOtherContextDirect.destroy()
            self._fOtherContextDirect = None
            
        if self._fNetChannel:
            self._fNetChannel.Close()
            self._fNetChannel = None
        
        if self._fSocketWaiter:
            self._fSocketWaiter.destroy()
            self._fSocketWaiter = None

    def SetDefaultExceptionCallback(self, callback):
        """ Set exception callback to be used if no exception callback
        is passed to the operation invocation. """
        
        self._fDefaultExceptionCallback = callback
        
    def SetNetChannel(self, net_channel, add_to_mainloop=True):
        """ Set the channel used to communicate with remote process.  If
        add_to_mainloop is true, set things up so calls will be dispatched
        from mainloop """
        
        self._fNetChannel = net_channel
        
        self.emit('connected', self)
        
        for to_send in self._fPendingOutgoingCalls:
            self._fNetChannel.SendPacket(to_send)
        self._fPendingOutgoingCalls = []
        
        if add_to_mainloop:
            self._SetupDataAvailableCheck()
                
    def SetLocalPrimaryObject(self, obj):
        """ Sets the primary object; the primary object's obj_id is always
        0 so methods on it may be invoked without first obtaining its id """
        
        self.fPrimaryObject = obj
        if obj is not None:
            self.fSerializer.AddNetworkAccessibleObject(obj)
    
    def GetPrimaryRemoteProxy(self):
        """ Returns proxy for primary object in remote process. """
        
        return self.fSerializer.GetProxy(hex(0))
        
    def SetOtherContextDirect(self, other):
        """ Set the other end of the context if it's in the same process to
        allow direct calls. """
        
        self._fOtherContextDirect = other

    def SetDefaultExceptionCallback(self, callback, proxy=None):
        
        if proxy is None:
            self._fDefaultExceptionCallback = callback
        else:
            proxy._fDefaultExceptionCallback = callback

    #----------------------------------------------------------------------
    def WaitForCommand(self, blocking=1, allow_recursive=0):

        ## If connection to client is gone, check whether there are any
        ## pending connection requests and try to reconnect if so
        ## If we cannot connect, then just keep going at any bps or exception
        #if self.__fChannel is None or self.__fChannel.IsClosed():
          #self.__CheckForConnectionRequests()
          #if self.__fServer != None and not self.__fServer.IsRunning():
            #self.__fServer.Continue()
          #return 0
    
        ## Check for recursive call (should not occur)
        #if not allow_recursive and self._fWaitingForCommand:
          #self.fErr.out('recursive call to WaitForCommand')
          #if self.fErr.AnyOutputFiles():
            #tbf = traceback.format_stack()
            #for item in tbf:
              #self.fErr.out(item)
          #return 0
        
        @contextlib.contextmanager
        def incremented_waiting_for_command():
            self._fWaitingForCommand += 1
            try:
                yield
            finally:
                self._fWaitingForCommand -= 1

        with incremented_waiting_for_command():
            executed = False
            
            self._LogActivity('in WaitFor')
  
            # Get next client command, if any
            try:
      
                # Wait for next client command
                timeout = (0 if blocking else None)
                self._LogActivity("Calling WaitForPacket(timeout=%r)", timeout)
                if self._fNetChannel is not None and self._fNetChannel.WaitForPacket(timeout):
                    packetdata = self._fNetChannel.ReceivePacket(blocking=1)
                else:
                    packetdata = None
      
                # Unpack and execute packet if valid and session hasn't ended
                if packetdata != None and self._fNetChannel != None and \
                   not self._fNetChannel.IsClosed():
                    self.ProcessIncomingPacket(packetdata)
                    executed = True
                    
            # Close connection on any error in the debugger code itself
            
            except Exception, exc:

                self._LogActivity('Exception occurred in WaitForCommand')
                self._LogActivity('Returning')
                
                if (isinstance(exc, channel.CNetworkChannel.CSocketError) 
                      and exc.id == errno.EPIPE):
                    logging.debug("Pipe broken exception is WaitForCommand")
                else:
                    logging.exception('Exception occurred in WaitForCommand')

                self.Shutdown(immediate=True)
                # Print exception info
                #self.fErr.out("EXCEPTION in WaitForCommand()")
                #for err in dbgutils.GetExceptionList():
                  #self.fErr.out(err)
        
                ## Send program stopped message back to client
                #self.ProgramStopped(self, [], -1, None, None, kFatalStop)
        
                ## Close channel and quit program
                #self.__CloseChannel()
                ## XXX Probably shouldn't call quit if externally launched
                #self.__fServer.Quit()
              
        
  #   self.fErr.out("exiting netserver.WaitForCommand")

        
    def SendMethodCall(self, obj_id, method_name, args, kw, callback, exception,
                       callback_if_destroyed):
        """ Send message to execute named method in the remote context. """
        
        
        request_id = self._fNextActiveRequestID
        self._fNextActiveRequestID += 1
        
        if callback is not None or exception is not None:
            self._StoreCallbacks(request_id, obj_id, callback, exception, 
                                 callback_if_destroyed)
            rtype = kNamedMethodCall
        else:
            rtype = kNoResultNamedMethodCall
            
        to_send = (rtype, request_id, obj_id, method_name, args, kw)
        
        if self._fNetChannel:
            try:
                self._fNetChannel.SendPacket(to_send)
            except Exception as exc:
                if exception is not None:
                    exception(exc)
                else:
                    logging.exception("No exception callback defined")
        else:
            raise ValueError()
            self._fPendingOutgoingCalls.append(to_send)
            
    def _SendDataTuple(self, data_tuple, ignore_exceptions=False):
        
        try:
            self._fNetChannel.SendPacket(data_tuple)
        except Exception:
            if not ignore_exceptions:
                raise
        
    def ProcessIncomingPacket(self, data_tuple):
        """ Process a single incoming packet.  ReceivePacket will return a
        tuple; the first element is the type of request. """
        
        rtype = data_tuple[0]
        if rtype in (kNamedMethodCall, kNoResultNamedMethodCall):
            rtype, request_id, obj_id, method_name, args, kw = data_tuple
            if obj_id in self._fDiscardCallsObjectIDs:
                return
            result, exc_info = self.CallMethod(obj_id, method_name, args, kw)
            if rtype == kNamedMethodCall:
                if exc_info:
                    import traceback
                    etype, exc, tb = exc_info
                    extracted = traceback.extract_tb(tb)
                    exc_info = exc, extracted
                to_send = (kMethodCallResult, request_id, result, exc_info)
                self._fNetChannel.SendPacket(to_send)
                
        elif rtype == kMethodCallResult:
            rtype, request_id, result, exc_info = data_tuple
            callback, exception = self._PopCallbacks(request_id)
            if exc_info is None:
                if callback:
                    callback(result)
                elif result is not None:
                    logging.debug('Non-None result received for'
                                  + ' request with no callback')
            else:
                if not exception:
                    exception = self._fDefaultExceptionCallback
                    
                if exception:
                    exc, tb_extact = exc_info
                    try:
                        exc._remote_traceback_extract
                    except AttributeError:
                        exc._remote_traceback_extract = tb_extact
                        
                    try:
                        raise exc
                    except Exception:
                        exception(exc)
                else:
                    logging.exception("No exception callback defined", 
                                      exc_info=exc_info)
                    
        elif rtype == kShutdown:
            self.Shutdown()
            
        print self.fPrimaryObject, 'End ProcessIncomingPacket'
                        
    def CallMethod(self, obj_id, method_name, args, kw):
        """ Call down to method on the local object w/ obj_id.  Returns 
        (return_value, exc_info) as a pair; return_value will be None if an
        exception occurred and exc_info will be None if an exception
        did not occur. """
        
        method = None
        retval = None
        exc_info = None
        try:
            obj = self.fSerializer.GetLocalObject(obj_id)
            method = getattr(obj, method_name)
            
            self._LogActivity('Calling method; method = %r', method)
            
            try:
                retval = method(*args, **kw)
            except:
                exc_info = sys.exc_info()
                logging.exception("Exception raised when calling: %r" % method,
                                  exc_info=exc_info)
            else:
                exc_info = None
                
            self._LogActivity('Returned from method')
                
        except:
            exc_info = sys.exc_info()
            logging.exception("Exception in CallMethod in call to: %r" % (method,),
                              exc_info=exc_info)
            
        return retval, exc_info
    
    def _SetupDataAvailableCheck(self):
        """ Setup callbacks to react to data available on the net channel.  A qt
        socket notifier is used and an timeout is also added.  The timeout is needed
        because the socket notifier as proven to be unreliable on win32 """
        
        if not self._fSocketWaiter:
            self._fSocketWaiter = _CWaitForData(self._fNetChannel.GetSocket(),
                                                self._CB_CheckForIncomingData)
    
    def _CB_CheckForIncomingData(self):
        """ Check for incoming data.  Gets called periodically and also in
        response to a data available notification """
    
        self._LogActivity("CheckForIncoming; waiting = %s", self._fWaitingForCommand)
        if not self._fWaitingForCommand:
            self.WaitForCommand(blocking=False)
            
    def _StoreCallbacks(self, request_id, obj_id, callback, exception, 
                        callback_if_destroyed):
        """ Store the callbacks in dict for active requests. """
        
        self._fCallbacksForActiveRequests[request_id] = (obj_id, callback, exception,
                                                         callback_if_destroyed)
        
    def _PopCallbacks(self, request_id):
        """ Pops callbacks from dict for active requests.  Checks to see if
        proxy has been destroyed. Either callback can be None. """
        
        item = self._fCallbacksForActiveRequests.pop(request_id)
        obj_id, callback, exception, callback_if_destroyed = item
        if not callback_if_destroyed and obj_id not in self.fSerializer._fActiveProxies:
            logging.debug('Call result received after proxy is destroyed')
            return None, None
        
        return callback, exception

class _CWaitForData(destroyable.CDestroyable):
    """ Arranges for a callback to be called when data is available on a 
    socket.  Currently also calls the callback periodically because the
    socket notifier on win32 isn't 100% reliable.  Will not call the 
    callback recursively. """
    
    kLogActivity = False
    
    _kDataAvailableCheckWaitPeriod = 250

    def __init__(self, sock, callback):
        
        destroyable.CDestroyable.__init__(self)
        
        self._fCallback = callback
        self._fInCallback = False
        
        mainloop = GetMainloop()
        mainloop.timeout_add_while_alive(self._kDataAvailableCheckWaitPeriod,
                                         self._CB_InvokeCallback, self)
            
        self._fSocketNotifier = mainloop.add_socket_watch(
            sock, self._CB_InvokeCallback, readable=True)
        
    def _destroy_impl(self):
        
        Destroy(self._fSocketNotifier)
        mainloop = GetMainloop()
        # XXX should remove
        #mainloop.timeout_remove(self._kDataAvailableCheckWaitPeriod,
        #                                 self._CB_InvokeCallback, self)
        
    def _CB_InvokeCallback(self):
        
        if self.destroyed():
            return False
        
        if self._fInCallback:
            return True
        
        if self.kLogActivity:
            print 'invoking callback in _CB_InvokeCallback:', self._fCallback
        
        self._fInCallback = True
        try:
            self._fCallback()
        finally:
            if not self.destroyed():
                self._fInCallback = False

        if self.kLogActivity:
            if not self.destroyed():
                print 'returning from callback in _CB_InvokeCallback:', self._fCallback
            else:
                print 'returning from callback in _CB_InvokeCallback: self is destroyed'
            sys.stdout.flush()
    
        return not self.destroyed()
        
class CProxy(destroyable.CDestroyable):
    """ A proxy object to call method on objects in the remote context. """
    
    def __init__(self, context, obj_id):
        
        super(CProxy, self).__init__()
        
        self._fContext = context
        self._fObjID = obj_id
        self._fDefaultExceptionCallback = None
        
    def _destroy_impl(self):
        
        exception = self._fDefaultExceptionCallback
        if exception is None:
            exception = self._fContext._fDefaultExceptionCallback
        
        self._fContext.SendMethodCall(self._fObjID, 'destroy', args=(), kw={},
                                      callback=None, exception=exception,
                                      callback_if_destroyed=False)
        self._fContext.fSerializer.RemoveActiveProxy(self)
        
    def __getattribute__(self, name):
        """ Called on all attribute access.  If name does not exist on self,
        it is assumed to be a method; all method calls must include a 
        callback and exception keyword arguments.  If not None, these
        are called when the operation / method completes with a return
        value or an exception, respectively. """
        
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            pass
    
        context = self._fContext
        obj_id = self._fObjID

        def wrap(*args, **kw):
          
            callback = kw.pop('callback')
            if 'exception' in kw:
                exception = kw.pop('exception')
            else:
                exception = self._fDefaultExceptionCallback
                if exception is None:
                    exception = context._fDefaultExceptionCallback
                    if exception is None:
                        raise TypeError, 'Exception callback must be non None'
            callback_if_destroyed = kw.pop('callback_if_destroyed', False)

            context.SendMethodCall(obj_id, name, args, kw, callback, exception,
                                   callback_if_destroyed)
        
        return wrap

def CreateInProcessRemoteOperations(primary_obj, use_channels=True):
    """ Create a remote operations context in the local address space, using idles
    to perform each call asynchronously. """
    
    remote_context = CContext()
    local_context = CContext()
    
    if use_channels:
        pair = SocketPair()
    
        local_channel = channel.CNetworkChannel(5.0, serializer=local_context.fSerializer)
        local_channel._UseSocket(pair[0])
    
        remote_channel = channel.CNetworkChannel(5.0, serializer=remote_context.fSerializer)
        remote_channel._UseSocket(pair[1])
        
        remote_channel._WriteInitiationPacket()
        local_channel._ReadInitiationPacket()
        local_channel._WriteInitiationPacket()
        remote_channel._ReadInitiationPacket()
        
        local_context.SetNetChannel(local_channel)
        remote_context.SetNetChannel(remote_channel)

    remote_context.SetLocalPrimaryObject(primary_obj)
    return local_context, remote_context
    
class CSeparateProcessForRemoteOp(destroyable.CDestroyable):
    """ Handles launching, setting up communication with, and terminating
    another process for remote operations """
    
    def __init__(self, argv, env=None, cwd=None):
        
        if spawn is None:
            raise RuntimeError("The spawn module is needed for launching a process")
        
        destroyable.CDestroyable.__init__(self)
        
        self.fLocalCtx = CContext()
        self.fLocalCtx.connect('destroy', self.destroy)

        self.fConnected = False
        self._fStartFailed = False
        
        self._fArgv = argv
        self._fEnv = env
        self._fCwd = cwd
        self.fProcess = spawn.CChildProcess(argv, env=self._fEnv, 
                                            child_pwd=self._fCwd)
        
        self._fUseSocketPair = (sys.platform != 'win32')
        
        self._fLocalChannel = None
        self._fConnectionWaiter = None
        
    def _destroy_impl(self):
        
        self.fLocalCtx.destroy()
        self.KillProcess()
        
    def KillProcess(self):
        """ Kills the process and cleans up """
        
        if self.fProcess:
            self.fProcess.Kill()
            self.fProcess.destroy()
            self.fProcess = None

        if self._fConnectionWaiter:
            Destroy(self._fConnectionWaiter)
            self._fConnectionWaiter = None
            
    def LaunchProcess(self):
        """ Launch the process; starts listening if using the connect back 
        style of connecting. """
        
        args = list(self._fArgv)
        keep_open_fds = []
        pair = None
        if self._fUseSocketPair:
            pair = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
            for end in pair:
                # XXX Uncomment the following to set both ends to be nonblocking
                # end.setblocking(False)

                # On OS X at least, a large buffer is needed so writes on
                # a nonblocking socket don't time out when writing a large
                # packet (10+ MB)
                bufsize = 1024 * 1024
                end.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, bufsize)
                end.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, bufsize)

            self._fLocalChannel = channel.CNetworkChannel(5.0, serializer=self.fLocalCtx.fSerializer)
            
            self._fLocalChannel._UseSocket(pair[0])
            socket_fd = pair[1].fileno()
            args.append('--socket-fd=' + str(socket_fd))
            keep_open_fds.append(socket_fd)
        else:
            self._fLocalChannel = channel.CNetworkChannel(5.0, serializer=self.fLocalCtx.fSerializer)

            self._fLocalChannel.Listen([kLocalHost], kLocalHost, 0, blocking=False)
            port = self._fLocalChannel.GetSocket().getsockname()[1]
            
            args.append('--connect-back=%s' % port)
            
        self.fProcess.SetStartParams(args=args, keep_open_fds=keep_open_fds)
        self.fProcess.connect_while_alive(
            'received-output', self.__CB_ReceivedOutput, self)
        self.fProcess.connect_while_alive(
            'terminated', self.__CB_Terminated, self)
        self.fProcess.connect('start-failed', self.__CB_StartFailed)

        try:
            self.fProcess.Start()
        except:
            if self.destroyed():
                return
            self._fStartFailed = True
        finally:
            if pair:
                pair[1].close()

        if self._fStartFailed:
            return

        self._fConnectionWaiter = _CWaitForData(self._fLocalChannel.GetSocket(),
                                                self._CB_CheckForConnection)

        if pair:
            self._fLocalChannel._WriteInitiationPacket()
        
    def _CB_CheckForConnection(self):
        """ Callback to check for a connection. """
        
        if self._fUseSocketPair:
            try:
                initiated = self._fLocalChannel._ReadInitiationPacket(blocking=0)
            except channel.CNetworkChannel.CSocketError:
                initiated = False
                
            if initiated:
                self.fConnected = True
            else:
                runnning = self.fProcess.GetRunning(check_for_output=True)
                if not runnning:
                    self.__CB_StartFailed(self.fProcess, None)
        else:
            self._fLocalChannel.Listen([kLocalHost], kLocalHost, 0, blocking=False)
            if not self._fLocalChannel.fListening:
                self.fConnected = True
                
        if self.fConnected:
            # Qt workaround: only 1 notifier for a socket is allowed at a
            # a time (at least on OS X) so delete the connection waiter before
            # setting the channel
            if self._fConnectionWaiter:
                Destroy(self._fConnectionWaiter)
                self._fConnectionWaiter = None

            self.fLocalCtx.SetNetChannel(self._fLocalChannel)
            
    def __CB_StartFailed(self, process, exc):
        
        self._fStartFailed = True
        
        self.__CB_Terminated(process)
        
    def __CB_ReceivedOutput(self, process, text):
        
        print text
        
    def __CB_Terminated(self, *args):
        
        if self._fLocalChannel:
            self._fLocalChannel.Close()
            self._fLocalChannel = None

        if self._fConnectionWaiter:
            Destroy(self._fConnectionWaiter)
            self._fConnectionWaiter = None
            
        if not self.fLocalCtx.destroyed():
            def idle(obj=self.fLocalCtx):
                wgtk.Destroy(obj)
            mainloop.GetMainloop().timeout_add_while_alive(0, idle, None)
        
def CreateSeparateProcess(py_file, args=(), env=None, py_interp=sys.executable):
    
    argv = [py_interp, py_file] + list(args)
    process = CSeparateProcessForRemoteOp(argv, env=env)
    process.LaunchProcess()
    return process.fLocalCtx
    
        
def PopArgEqual(args, name):
    """ Find first arg starting with name in args, remove it from args, and return
    anything after the name.  None is returned if no such arg is found. """
    
    for i, a in enumerate(args):
        if a.startswith(name):
            del args[i]
            return a[len(name):]
        
    return None

def TryInt(s):
    """ Convert to int; return None if there is an exception """
    
    try:
        return int(s)
    except Exception:
        return None
    
def Main(primary_obj, args=None):
    """ Function to be called in a remote process to setup a context to serve the 
    given primary object, set up a channel to the other process, and start a main
    loop via wgtk.main() to process requests.  Will exit the mainloop when a
    terminate request is recieved. """
    
    if args is None:
        args = list(sys.argv[1:])
        
    socket_fd = PopArgEqual(args, '--socket-fd=')
    connect_back = PopArgEqual(args, '--connect-back=')

    if socket_fd:
        socket_fd = TryInt(socket_fd)
    if connect_back:
        connect_back = TryInt(connect_back)
            
    if not (socket_fd or connect_back):
        import logging
        logging.critical('No way to connect to the other process')
        return -1

    local_context = CContext()
    local_context.SetLocalPrimaryObject(primary_obj)
    
    if socket_fd:
        s = socket.fromfd(socket_fd, socket.AF_UNIX, socket.SOCK_STREAM)
        local_channel = channel.CNetworkChannel(5.0, serializer=local_context.fSerializer)
        local_channel._UseSocket(s)
        local_context.SetNetChannel(local_channel)
    
        local_channel._ReadInitiationPacket()
        local_channel._WriteInitiationPacket()
    elif connect_back:
        local_channel = channel.CNetworkChannel(5.0, serializer=local_context.fSerializer)
        local_channel.Connect(kLocalHost, connect_back)
        local_context.SetNetChannel(local_channel)
        
    def exit_on_shutdown(ctx):
        GetMainloop().mainquit()
        
    local_context.connect('shutdown', exit_on_shutdown)
        
    mainloop.GetMainloop().main()
    